import os
import json
from pathlib import Path
import pytest
from unittest.mock import patch, MagicMock
from helpers.Project import Project

test_root = str(Path(__file__).parent.parent.parent / Path("workspace") / Path("gpt-pilot-test"))

def create_project():
    project = Project({
        'app_id': 'test-project',
        'name': 'TestProject',
        'app_type': ''
    })
    project.set_root_path(test_root)
    project.app = 'test'
    project.current_step = 'test'
    return project


class TestProject:
    @pytest.mark.parametrize('file_path, file_name, expected', [
        ('file.txt', 'file.txt', f'{test_root}/file.txt'),
        ('', 'file.txt', f'{test_root}/file.txt'),
        ('path/', 'file.txt', f'{test_root}/path/file.txt'),
        ('path/to/', 'file.txt', f'{test_root}/path/to/file.txt'),
        ('path/to/file.txt', 'file.txt', f'{test_root}/path/to/file.txt'),
        ('./path/to/file.txt', 'to/file.txt', f'{test_root}/path/to/file.txt'),
        ('./package.json', 'package.json', f'{test_root}/package.json'),
    ])
    def test_get_full_path(self, file_path, file_name, expected):
        # Given
        project = create_project()

        # When
        relative_path, absolute_path = project.get_full_file_path(file_path, file_name)

        # Then
        assert absolute_path == str(Path(expected))

    @pytest.mark.parametrize(
        ("file_path", "file_name", "expected_path", "expected_absolute_path"), [
        ('', '', '/', f'{test_root}/'),
        ('', '.', '/', f'{test_root}/'),
        ('', '.env', '/', f'{test_root}/.env'),
        ('', '~/', '/', f'{test_root}/'),
        ('', f'{test_root}/', '/', f'{test_root}/'),
        ('', f'{test_root}/folder1', '/folder1', f'{test_root}/folder1/'),
        ('', f'{test_root}/Folder With Space/', '/Folder With Space', f'{test_root}/Folder With Space/'),
        ('', 'server.js', '/', f'{test_root}/server.js'),
        ('', 'folder1', '/folder1', f'{test_root}/folder1/'),
        ('', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('', f'{test_root}/Folder With Space/server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        ('', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('', './folder1/server.js', '/folder1', f'{test_root}/folder1/server.js'),
        ('.', '', '/', f'{test_root}/'),
        ('.', '.', '/', f'{test_root}/'),
        ('.', '.env', '/', f'{test_root}/.env'),
        ('.', '~/', '/', f'{test_root}/'),
        ('.', f'{test_root}/', '/', f'{test_root}/'),
        ('.', f'{test_root}/folder1', '/folder1', f'{test_root}/folder1/'),
        ('.', f'{test_root}/Folder With Space/', '/Folder With Space', f'{test_root}/Folder With Space/'),
        ('.', 'server.js', '/', f'{test_root}/server.js'),
        ('.', 'folder1', '/folder1', f'{test_root}/folder1/'),
        ('.', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('.', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('.', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('.', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('.', f'{test_root}/Folder With Space/server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        ('.', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('.', './folder1/server.js', '/folder1', f'{test_root}/folder1/server.js'),
        ('.env', '', '/', f'{test_root}/.env'),
        ('.env', '.', '/', f'{test_root}/.env'),
        ('.env', '.env', '/', f'{test_root}/.env'),
        ('.env', '~/', '/', f'{test_root}/.env'),
        ('.env', f'{test_root}/', '/', f'{test_root}/.env'),
        ('.env', f'{test_root}/folder1', '/folder1', f'{test_root}/folder1/.env'),
        ('.env', f'{test_root}/Folder With Space/', '/Folder With Space', f'{test_root}/Folder With Space/.env'),
        ('.env', 'server.js', '/', f'{test_root}/server.js'),
        ('.env', 'folder1', '/folder1', f'{test_root}/folder1/.env'),
        ('.env', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/.env'),
        ('.env', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/.env'),
        ('.env', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('.env', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('.env', f'{test_root}/Folder With Space/server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        ('.env', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('.env', './folder1/server.js', '/folder1', f'{test_root}/folder1/server.js'),
        ('~/', '', '/', f'{test_root}/'),
        ('~/', '.', '/', f'{test_root}/'),
        ('~/', '.env', '/', f'{test_root}/.env'),
        ('~/', '~/', '/', f'{test_root}/'),
        ('~/', f'{test_root}/', '/', f'{test_root}/'),
        ('~/', f'{test_root}/folder1', '/folder1', f'{test_root}/folder1/'),
        ('~/', f'{test_root}/Folder With Space/', '/Folder With Space', f'{test_root}/Folder With Space/'),
        ('~/', 'server.js', '/', f'{test_root}/server.js'),
        ('~/', 'folder1', '/folder1', f'{test_root}/folder1/'),
        ('~/', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('~/', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('~/', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/', f'{test_root}/Folder With Space/server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        ('~/', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/', './folder1/server.js', '/folder1', f'{test_root}/folder1/server.js'),
        (f'{test_root}/', '', '/', f'{test_root}/'),
        (f'{test_root}/', '.', '/', f'{test_root}/'),
        (f'{test_root}/', '.env', '/', f'{test_root}/.env'),
        (f'{test_root}/', '~/', '/', f'{test_root}/'),
        (f'{test_root}/', f'{test_root}/', '/', f'{test_root}/'),
        (f'{test_root}/', f'{test_root}/folder1', '/folder1', f'{test_root}/folder1/'),
        (f'{test_root}/', f'{test_root}/Folder With Space/', '/Folder With Space', f'{test_root}/Folder With Space/'),
        (f'{test_root}/', 'server.js', '/', f'{test_root}/server.js'),
        (f'{test_root}/', 'folder1', '/folder1', f'{test_root}/folder1/'),
        (f'{test_root}/', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        (f'{test_root}/', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        (f'{test_root}/', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/', f'{test_root}/Folder With Space/server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/', './folder1/server.js', '/folder1', f'{test_root}/folder1/server.js'),
        (f'{test_root}/folder1', '', '/folder1', f'{test_root}/folder1/'),
        (f'{test_root}/folder1', '.', '/folder1', f'{test_root}/folder1/'),
        (f'{test_root}/folder1', '.env', '/folder1', f'{test_root}/folder1/.env'),
        (f'{test_root}/folder1', '~/', '/folder1', f'{test_root}/folder1/'),
        (f'{test_root}/folder1', f'{test_root}/', '/folder1', f'{test_root}/folder1/'),
        (f'{test_root}/folder1', f'{test_root}/folder1', '/folder1', f'{test_root}/folder1/'),
        (f'{test_root}/folder1', f'{test_root}/Folder With Space/', '/folder1/Folder With Space', f'{test_root}/folder1/Folder With Space/'),
        (f'{test_root}/folder1', 'server.js', '/folder1', f'{test_root}/folder1/server.js'),
        (f'{test_root}/folder1', 'folder1', '/folder1', f'{test_root}/folder1/'),
        (f'{test_root}/folder1', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        (f'{test_root}/folder1', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        (f'{test_root}/folder1', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1', f'{test_root}/Folder With Space/server.js', '/folder1/Folder With Space', f'{test_root}/folder1/Folder With Space/server.js'),
        (f'{test_root}/folder1', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1', './folder1/server.js', '/folder1', f'{test_root}/folder1/server.js'),
        (f'{test_root}/Folder With Space/', '', '/Folder With Space', f'{test_root}/Folder With Space/'),
        (f'{test_root}/Folder With Space/', '.', '/Folder With Space', f'{test_root}/Folder With Space/'),
        (f'{test_root}/Folder With Space/', '.env', '/Folder With Space', f'{test_root}/Folder With Space/.env'),
        (f'{test_root}/Folder With Space/', '~/', '/Folder With Space', f'{test_root}/Folder With Space/'),
        (f'{test_root}/Folder With Space/', f'{test_root}/', '/Folder With Space', f'{test_root}/Folder With Space/'),
        (f'{test_root}/Folder With Space/', f'{test_root}/folder1', '/Folder With Space/folder1', f'{test_root}/Folder With Space/folder1/'),
        (f'{test_root}/Folder With Space/', f'{test_root}/Folder With Space/', '/Folder With Space', f'{test_root}/Folder With Space/'),
        (f'{test_root}/Folder With Space/', 'server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/', 'folder1', '/Folder With Space/folder1', f'{test_root}/Folder With Space/folder1/'),
        (f'{test_root}/Folder With Space/', 'folder1/folder2', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/'),
        (f'{test_root}/Folder With Space/', 'folder1/folder2/', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/'),
        (f'{test_root}/Folder With Space/', 'folder1/folder2/server.js', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/server.js'),
        (f'{test_root}/Folder With Space/', f'{test_root}/folder1/folder2/server.js', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/server.js'),
        (f'{test_root}/Folder With Space/', f'{test_root}/Folder With Space/server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/', '~/folder1/folder2/server.js', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/server.js'),
        (f'{test_root}/Folder With Space/', './folder1/server.js', '/Folder With Space/folder1', f'{test_root}/Folder With Space/folder1/server.js'),
        ('server.js', '', '/', f'{test_root}/server.js'),
        ('server.js', '.', '/', f'{test_root}/server.js'),
        ('server.js', '.env', '/', f'{test_root}/.env'),
        ('server.js', '~/', '/', f'{test_root}/server.js'),
        ('server.js', f'{test_root}/', '/', f'{test_root}/server.js'),
        ('server.js', f'{test_root}/folder1', '/folder1', f'{test_root}/folder1/server.js'),
        ('server.js', f'{test_root}/Folder With Space/', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        ('server.js', 'server.js', '/', f'{test_root}/server.js'),
        ('server.js', 'folder1', '/folder1', f'{test_root}/folder1/server.js'),
        ('server.js', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('server.js', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('server.js', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('server.js', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('server.js', f'{test_root}/Folder With Space/server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        ('server.js', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('server.js', './folder1/server.js', '/folder1', f'{test_root}/folder1/server.js'),
        ('folder1', '', '/folder1', f'{test_root}/folder1/'),
        ('folder1', '.', '/folder1', f'{test_root}/folder1/'),
        ('folder1', '.env', '/folder1', f'{test_root}/folder1/.env'),
        ('folder1', '~/', '/folder1', f'{test_root}/folder1/'),
        ('folder1', f'{test_root}/', '/folder1', f'{test_root}/folder1/'),
        ('folder1', f'{test_root}/folder1', '/folder1', f'{test_root}/folder1/'),
        ('folder1', f'{test_root}/Folder With Space/', '/folder1/Folder With Space', f'{test_root}/folder1/Folder With Space/'),
        ('folder1', 'server.js', '/folder1', f'{test_root}/folder1/server.js'),
        ('folder1', 'folder1', '/folder1', f'{test_root}/folder1/'),
        ('folder1', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1', f'{test_root}/Folder With Space/server.js', '/folder1/Folder With Space', f'{test_root}/folder1/Folder With Space/server.js'),
        ('folder1', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1', './folder1/server.js', '/folder1', f'{test_root}/folder1/server.js'),
        ('folder1/folder2', '', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2', '.', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2', '.env', '/folder1/folder2', f'{test_root}/folder1/folder2/.env'),
        ('folder1/folder2', '~/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2', f'{test_root}/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2', f'{test_root}/folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2', f'{test_root}/Folder With Space/', '/folder1/folder2/Folder With Space', f'{test_root}/folder1/folder2/Folder With Space/'),
        ('folder1/folder2', 'server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2', 'folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2', f'{test_root}/Folder With Space/server.js', '/folder1/folder2/Folder With Space', f'{test_root}/folder1/folder2/Folder With Space/server.js'),
        ('folder1/folder2', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2', './folder1/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/', '', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2/', '.', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2/', '.env', '/folder1/folder2', f'{test_root}/folder1/folder2/.env'),
        ('folder1/folder2/', '~/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2/', f'{test_root}/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2/', f'{test_root}/folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2/', f'{test_root}/Folder With Space/', '/folder1/folder2/Folder With Space', f'{test_root}/folder1/folder2/Folder With Space/'),
        ('folder1/folder2/', 'server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/', 'folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2/', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2/', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/'),
        ('folder1/folder2/', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/', f'{test_root}/Folder With Space/server.js', '/folder1/folder2/Folder With Space', f'{test_root}/folder1/folder2/Folder With Space/server.js'),
        ('folder1/folder2/', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/', './folder1/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', '', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', '.', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', '.env', '/folder1/folder2', f'{test_root}/folder1/folder2/.env'),
        ('folder1/folder2/server.js', '~/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', f'{test_root}/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', f'{test_root}/folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', f'{test_root}/Folder With Space/', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', 'server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', 'folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', f'{test_root}/Folder With Space/server.js', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('folder1/folder2/server.js', './folder1/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', '', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', '.', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', '.env', '/folder1/folder2', f'{test_root}/folder1/folder2/.env'),
        (f'{test_root}/folder1/folder2/server.js', '~/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', f'{test_root}/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', f'{test_root}/folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', f'{test_root}/Folder With Space/', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', 'server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', 'folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', f'{test_root}/Folder With Space/server.js', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/folder1/folder2/server.js', './folder1/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        (f'{test_root}/Folder With Space/server.js', '', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', '.', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', '.env', '/Folder With Space', f'{test_root}/Folder With Space/.env'),
        (f'{test_root}/Folder With Space/server.js', '~/', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', f'{test_root}/', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', f'{test_root}/folder1', '/folder1/Folder With Space', f'{test_root}/folder1/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', f'{test_root}/Folder With Space/', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', 'server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', 'folder1', '/folder1/Folder With Space', f'{test_root}/folder1/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', 'folder1/folder2', '/folder1/folder2/Folder With Space', f'{test_root}/folder1/folder2/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', 'folder1/folder2/', '/folder1/folder2/Folder With Space', f'{test_root}/folder1/folder2/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', 'folder1/folder2/server.js', '/folder1/folder2/Folder With Space', f'{test_root}/folder1/folder2/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2/Folder With Space', f'{test_root}/folder1/folder2/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', f'{test_root}/Folder With Space/server.js', '/Folder With Space', f'{test_root}/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', '~/folder1/folder2/server.js', '/folder1/folder2/Folder With Space', f'{test_root}/folder1/folder2/Folder With Space/server.js'),
        (f'{test_root}/Folder With Space/server.js', './folder1/server.js', '/folder1/Folder With Space', f'{test_root}/folder1/Folder With Space/server.js'),
        ('~/folder1/folder2/server.js', '', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', '.', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', '.env', '/folder1/folder2', f'{test_root}/folder1/folder2/.env'),
        ('~/folder1/folder2/server.js', '~/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', f'{test_root}/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', f'{test_root}/folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', f'{test_root}/Folder With Space/', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', 'server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', 'folder1', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', f'{test_root}/Folder With Space/server.js', '/Folder With Space/folder1/folder2', f'{test_root}/Folder With Space/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('~/folder1/folder2/server.js', './folder1/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('./folder1/server.js', '', '/folder1', f'{test_root}/folder1/server.js'),
        ('./folder1/server.js', '.', '/folder1', f'{test_root}/folder1/server.js'),
        ('./folder1/server.js', '.env', '/folder1', f'{test_root}/folder1/.env'),
        ('./folder1/server.js', '~/', '/folder1', f'{test_root}/folder1/server.js'),
        ('./folder1/server.js', f'{test_root}/', '/folder1', f'{test_root}/folder1/server.js'),
        ('./folder1/server.js', f'{test_root}/folder1', '/folder1', f'{test_root}/folder1/server.js'),
        ('./folder1/server.js', f'{test_root}/Folder With Space/', '/Folder With Space/folder1', f'{test_root}/Folder With Space/folder1/server.js'),
        ('./folder1/server.js', 'server.js', '/folder1', f'{test_root}/folder1/server.js'),
        ('./folder1/server.js', 'folder1', '/folder1', f'{test_root}/folder1/server.js'),
        ('./folder1/server.js', 'folder1/folder2', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('./folder1/server.js', 'folder1/folder2/', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('./folder1/server.js', 'folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('./folder1/server.js', f'{test_root}/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('./folder1/server.js', f'{test_root}/Folder With Space/server.js', '/Folder With Space/folder1', f'{test_root}/Folder With Space/folder1/server.js'),
        ('./folder1/server.js', '~/folder1/folder2/server.js', '/folder1/folder2', f'{test_root}/folder1/folder2/server.js'),
        ('./folder1/server.js', './folder1/server.js', '/folder1', f'{test_root}/folder1/server.js'),

    ])
    def test_get_full_path_permutations(self, file_path, file_name, expected_path, expected_absolute_path):
        """
        Test many different permutations of file path/name combinations.
        """
        project = create_project()
        relative_path, absolute_path = project.get_full_file_path(file_path, file_name)
        # Normalize results before comparison, in case of Windows
        assert str(Path(relative_path)) == str(Path(expected_path))
        assert str(Path(absolute_path)) == str(Path(expected_absolute_path))

    @pytest.mark.parametrize('test_data', [
        {'name': 'package.json', 'path': 'package.json', 'saved_to': f'{test_root}/package.json'},
        {'name': 'package.json', 'path': '', 'saved_to': f'{test_root}/package.json'},
        {'name': 'package.json', 'path': '/', 'saved_to': f'{test_root}/package.json'},
        {'name': 'package.json', 'path': None, 'saved_to': f'{test_root}/package.json'},
        {'name': None, 'path': 'public/index.html', 'saved_to': f'{test_root}/public/index.html'},
        {'name': '', 'path': 'public/index.html', 'saved_to': f'{test_root}/public/index.html'},
        # TODO: Treatment of paths outside of the project workspace - https://github.com/Pythagora-io/gpt-pilot/issues/129
        {'name': '/etc/hosts.txt', 'path': None, 'saved_to': f'{test_root}/etc/hosts.txt'},
        # {'name': '.gitconfig', 'path': '~', 'saved_to': '~/.gitconfig'},
        # {'name': '.gitconfig', 'path': '~/.gitconfig', 'saved_to': '~/.gitconfig'},
        # {'name': 'gpt-pilot.log', 'path': '/temp/gpt-pilot.log', 'saved_to': '/temp/gpt-pilot.log'},
    ])
    @patch('helpers.Project.update_file')
    @patch('helpers.Project.File')
    def test_save_file(self, mock_file_insert, mock_update_file, test_data):
        # Given
        data = {'content': 'Hello World!'}
        if test_data['name'] is not None:
            data['name'] = str(Path(test_data['name']))
        if test_data['path'] is not None:
            data['path'] = str(Path(test_data['path']))

        project = create_project()

        # When
        project.save_file(data)

        # Then assert that update_file with the correct path
        expected_saved_to = str(Path(test_data['saved_to']))
        mock_update_file.assert_called_once_with(expected_saved_to, 'Hello World!', project=project)

        # Also assert that File.insert was called with the expected arguments
        # expected_file_data = {'app': project.app, 'path': test_data['path'], 'name': test_data['name'],
        #                       'full_path': expected_saved_to}
        # mock_file_insert.assert_called_once_with(**expected_file_data,

class TestProjectFileLists:
    def setup_method(self):
        # Given a project
        project = create_project()
        self.project = project
        project.set_root_path(os.path.join(os.path.dirname(__file__), '../../workspace/directory_tree'))
        project.project_description = 'Test Project'
        project.development_plan = [{
            'description': 'Test User Story',
            'user_review_goal': 'Test User Review Goal',
        }]

        # with directories including common.IGNORE_PATHS
        src = os.path.join(project.root_path, 'src')
        foo = os.path.join(project.root_path, 'src/foo')
        files_no_folders = os.path.join(foo, 'files_no_folders')
        os.makedirs(src, exist_ok=True)
        os.makedirs(foo, exist_ok=True)
        os.makedirs(foo + '/empty1', exist_ok=True)
        os.makedirs(foo + '/empty2', exist_ok=True)
        os.makedirs(files_no_folders, exist_ok=True)
        for dir in ['.git', '.idea', '.vscode', '__pycache__', 'node_modules', 'venv', 'dist', 'build']:
            os.makedirs(os.path.join(project.root_path, dir), exist_ok=True)

        # ...and files

        with open(os.path.join(project.root_path, 'package.json'), 'w') as file:
            json.dump({'name': 'test app'}, file, indent=2)
        for path in [
            os.path.join(src, 'main.js'),
            os.path.join(src, 'other.js'),
            os.path.join(foo, 'bar.js'),
            os.path.join(foo, 'fighters.js'),
            os.path.join(files_no_folders, 'file1.js'),
            os.path.join(files_no_folders, 'file2.js'),
        ]:
            with open(path, 'w') as file:
                file.write('console.log("Hello World!");')

        # and a non-empty .gpt-pilot directory
        project.dot_pilot_gpt.write_project(project)

    def test_get_directory_tree(self):
        # When
        tree = self.project.get_directory_tree()

        # Then we should not be including the .gpt-pilot directory or other ignored directories
        # print('\n' + tree)
        assert tree == '''
/
  /src
    /foo
      /empty1
      /empty2
      /files_no_folders: file1.js, file2.js
      bar.js, fighters.js
    main.js, other.js
  package.json
'''.lstrip()

    @patch('helpers.Project.DevelopmentSteps.get_or_create', return_value=('test', True))
    @patch('helpers.Project.File.get_or_create', return_value=('test', True))
    @patch('helpers.Project.FileSnapshot.get_or_create', return_value=(MagicMock(), True))
    def test_save_files_snapshot(self, mock_snap, mock_file, mock_step):
        # Given a snapshot of the files in the project

        # When we save the file snapshot
        self.project.save_files_snapshot('test')

        # Then the files should be saved to the project, but nothing from `.gpt-pilot/`
        assert mock_file.call_count == 7
        files = ['package.json', 'main.js', 'file1.js', 'file2.js', 'bar.js', 'fighters.js', 'other.js']
        for i in range(7):
            assert mock_file.call_args_list[i][1]['name'] in files



from peewee import ForeignKeyField, AutoField, TextField, IntegerField, CharField
from database.config import DATABASE_TYPE
from database.models.components.base_models import BaseModel
from database.models.app import App
from database.models.components.sqlite_middlewares import JSONField
from playhouse.postgres_ext import BinaryJSONField


class DevelopmentSteps(BaseModel):
    id = AutoField()  # This will serve as the primary key
    app = ForeignKeyField(App, on_delete='CASCADE')
    prompt_path = TextField(null=True)
    llm_req_num = IntegerField(null=True)
    token_limit_exception_raised = TextField(null=True)

    if DATABASE_TYPE == 'postgres':
        messages = BinaryJSONField(null=True)
        llm_response = BinaryJSONField(null=False)
        prompt_data = BinaryJSONField(null=True)
    else:
        messages = JSONField(null=True)  # Custom JSON field for SQLite
        llm_response = JSONField(null=False)  # Custom JSON field for SQLite
        prompt_data = JSONField(null=True)

    previous_step = ForeignKeyField('self', null=True, column_name='previous_step')
    high_level_step = CharField(null=True)

    class Meta:
        table_name = 'development_steps'
        indexes = (
            (('app', 'previous_step', 'high_level_step'), True),
        )



import json
import re
import subprocess
import uuid
from utils.style import color_yellow, color_yellow_bold

from database.database import get_saved_development_step, save_development_step, delete_all_subsequent_steps
from helpers.exceptions.TokenLimitError import TokenLimitError
from utils.function_calling import parse_agent_response, FunctionCallSet
from utils.llm_connection import create_gpt_chat_completion
from utils.utils import get_prompt, get_sys_message, capitalize_first_word_with_underscores
from logger.logger import logger
from prompts.prompts import ask_user
from const.llm import END_RESPONSE
from helpers.cli import running_processes


class AgentConvo:
    """
    Represents a conversation with an agent.

    Args:
        agent: An instance of the agent participating in the conversation.
    """

    def __init__(self, agent):
        # [{'role': 'system'|'user'|'assistant', 'content': ''}, ...]
        self.messages: list[dict] = []
        self.branches = {}
        self.log_to_user = True
        self.agent = agent
        self.high_level_step = self.agent.project.current_step

        # add system message
        system_message = get_sys_message(self.agent.role, self.agent.project.args)
        logger.info('\n>>>>>>>>>> System Prompt >>>>>>>>>>\n%s\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>',
                    system_message['content'])
        self.messages.append(system_message)

    def send_message(self, prompt_path=None, prompt_data=None, function_calls: FunctionCallSet = None, should_log_message=True):
        """
        Sends a message in the conversation.

        Args:
            prompt_path: The path to a prompt.
            prompt_data: Data associated with the prompt.
            function_calls: Optional function calls to be included in the message.
            should_log_message: Flag if final response should be logged.
        Returns:
            The response from the agent.
        """
        # craft message
        self.construct_and_add_message_from_prompt(prompt_path, prompt_data)

        # TODO: move this if block (and the other below) to Developer agent - https://github.com/Pythagora-io/gpt-pilot/issues/91#issuecomment-1751964079
        # check if we already have the LLM response saved
        if hasattr(self.agent, 'save_dev_steps') and self.agent.save_dev_steps:
            self.agent.project.llm_req_num += 1
        development_step = get_saved_development_step(self.agent.project)
        if development_step is not None and self.agent.project.skip_steps:
            # if we do, use it
            print(color_yellow(f'Restoring development step with id {development_step.id}'))
            self.agent.project.checkpoints['last_development_step'] = development_step
            self.agent.project.restore_files(development_step.id)
            response = development_step.llm_response
            self.messages = development_step.messages

            if self.agent.project.skip_until_dev_step and str(
                    development_step.id) == self.agent.project.skip_until_dev_step:
                self.agent.project.finish_loading()
                delete_all_subsequent_steps(self.agent.project)

                if 'delete_unrelated_steps' in self.agent.project.args and self.agent.project.args[
                    'delete_unrelated_steps']:
                    self.agent.project.delete_all_steps_except_current_branch()
            else:
                should_log_message = True

            if development_step.token_limit_exception_raised:
                raise TokenLimitError(development_step.token_limit_exception_raised)
        else:
            # if we don't, get the response from LLM
            try:
                self.replace_files()
                response = create_gpt_chat_completion(self.messages, self.high_level_step, self.agent.project,
                                                      function_calls=function_calls)
            except TokenLimitError as e:
                save_development_step(self.agent.project, prompt_path, prompt_data, self.messages, '', str(e))
                raise e

            # TODO: move this code to Developer agent - https://github.com/Pythagora-io/gpt-pilot/issues/91#issuecomment-1751964079
            if hasattr(self.agent, 'save_dev_steps') and self.agent.save_dev_steps:
                save_development_step(self.agent.project, prompt_path, prompt_data, self.messages, response)

        # TODO handle errors from OpenAI
        # It's complicated because calling functions are expecting different types of responses - string or tuple
        # https://github.com/Pythagora-io/gpt-pilot/issues/165 & #91
        if response == {}:
            logger.error('Aborting with "OpenAI API error happened"')
            raise Exception("OpenAI API error happened.")

        response = parse_agent_response(response, function_calls)
        message_content = self.format_message_content(response, function_calls)

        # TODO we need to specify the response when there is a function called
        # TODO maybe we can have a specific function that creates the GPT response from the function call
        logger.info('\n>>>>>>>>>> Assistant Prompt >>>>>>>>>>\n%s\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>',
                    message_content)
        self.messages.append({"role": "assistant", "content": message_content})
        if should_log_message:
            self.log_message(message_content)

        return response

    def format_message_content(self, response, function_calls):
        # TODO remove this once the database is set up properly
        if isinstance(response, str):
            return response
        else:
            # string_response = []
            # for key, value in response.items():
            #     string_response.append(f'# {key}')
            #
            #     if isinstance(value, list):
            #         if 'to_message' in function_calls:
            #             string_response.append(function_calls['to_message'](value))
            #         elif len(value) > 0 and isinstance(value[0], dict):
            #             string_response.extend([
            #                 f'##{i}\n' + array_of_objects_to_string(d)
            #                 for i, d in enumerate(value)
            #             ])
            #         else:
            #             string_response.extend(['- ' + r for r in value])
            #     else:
            #         string_response.append(str(value))
            #
            # return '\n'.join(string_response)
            return json.dumps(response)
        # TODO END

    def continuous_conversation(self, prompt_path, prompt_data, function_calls=None):
        """
        Conducts a continuous conversation with the agent.

        Args:
            prompt_path: The path to a prompt.
            prompt_data: Data associated with the prompt.
            function_calls: Optional function calls to be included in the conversation.

        Returns:
            List of accepted messages in the conversation.
        """
        self.log_to_user = False
        accepted_messages = []
        response = self.send_message(prompt_path, prompt_data, function_calls)

        # Continue conversation until GPT response equals END_RESPONSE
        while response != END_RESPONSE:
            user_message = ask_user(self.agent.project,
                                    'Do you want to add anything else? If not, just press ENTER.',
                                    hint=response,
                                    require_some_input=False)

            if user_message == "":
                accepted_messages.append(response)

            logger.info('\n>>>>>>>>>> User Message >>>>>>>>>>\n%s\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>', user_message)
            self.messages.append({"role": "user", "content": user_message})
            response = self.send_message(None, None, function_calls)

        self.log_to_user = True
        return accepted_messages

    def save_branch(self, branch_name=None):
        if branch_name is None:
            branch_name = str(uuid.uuid4())
        self.branches[branch_name] = self.messages.copy()
        return branch_name

    def load_branch(self, branch_name, reload_files=True):
        self.messages = self.branches[branch_name].copy()
        if reload_files:
            # TODO make this more flexible - with every message, save metadata so every time we load a branch, reconstruct all messages from scratch
            self.replace_files()

    def replace_files(self):
        files = self.agent.project.get_all_coded_files()
        for msg in self.messages:
            if msg['role'] == 'user':
                for file in files:
                    msg['content'] = self.replace_file_content(msg['content'], f"{file['path']}/{file['name']}", file['content'])

    @staticmethod
    def escape_specials(s):
        s = s.replace("\\", "\\\\")

        # List of sequences to preserve
        sequences_to_preserve = [
            # todo check if needed "\\\\",  # Backslash - note: probably not eg. paths on Windows
            "\\'",  # Single quote
            '\\"',  # Double quote
            # todo check if needed '\\a',  # ASCII Bell (BEL)
            # todo check if needed '\\b',  # ASCII Backspace (BS) - note: different from regex \b
            # todo check if needed '\\f',  # ASCII Formfeed (FF)
            '\\n',  # ASCII Linefeed (LF)
            # todo check if needed '\\r',  # ASCII Carriage Return (CR)
            '\\t',  # ASCII Horizontal Tab (TAB)
            # todo check if needed '\\v'  # ASCII Vertical Tab (VT)
        ]

        for seq in sequences_to_preserve:
            s = s.replace('\\\\' + seq[-1], seq)
        return s

    def replace_file_content(self, message, file_path, new_content):
        pattern = rf'\*\*{re.escape(file_path)}\*\*:\n```\n(.*?)\n```'

        # Escape special characters in new_content for the sake of regex replacement
        new_content_escaped = self.escape_specials(new_content)
        file_path_escaped = self.escape_specials(file_path)

        new_section_content = f'**{file_path_escaped}**\n```\n{new_content_escaped}\n```'

        updated_message, num_replacements = re.subn(pattern, new_section_content, message, flags=re.DOTALL)

        if num_replacements == 0:
            return message

        return updated_message

    def convo_length(self):
        return len([msg for msg in self.messages if msg['role'] != 'system'])

    def log_message(self, content):
        """
        Logs a message in the conversation.

        Args:
            content: The content of the message to be logged.
        """
        print_msg = capitalize_first_word_with_underscores(self.high_level_step)
        if self.log_to_user:
            if self.agent.project.checkpoints['last_development_step'] is not None:
                dev_step_msg = f'\nDev step {str(self.agent.project.checkpoints["last_development_step"])}\n'
                print(color_yellow_bold(dev_step_msg), end='')
                logger.info(dev_step_msg)
            print(f"\n{content}\n", type='local')
        logger.info(f"{print_msg}: {content}\n")

    def to_context_prompt(self):
        logger.info(f'to_context_prompt({self.agent.project.current_step})')

        # TODO: get dependencies & versions from the project (package.json, requirements.txt, pom.xml, etc.)
        # Ideally, the LLM could do this, and we update it on load & whenever the file changes
        # ...or LLM generates a script for `.gpt-pilot/get_dependencies` that we run
        # https://github.com/Pythagora-io/gpt-pilot/issues/189
        return get_prompt('development/context.prompt', {
            'directory_tree': self.agent.project.get_directory_tree(),
            'running_processes': running_processes,
        })

    def to_playground(self):
        # Internal function to help debugging in OpenAI Playground, not to be used in production
        with open('const/convert_to_playground_convo.js', 'r', encoding='utf-8') as file:
            content = file.read()
        process = subprocess.Popen('pbcopy', stdin=subprocess.PIPE)
        process.communicate(content.replace('{{messages}}', str(self.messages)).encode('utf-8'))

    def remove_last_x_messages(self, x):
        logger.info('removing last %d messages: %s', x, self.messages[-x:])
        self.messages = self.messages[:-x]

    def construct_and_add_message_from_prompt(self, prompt_path, prompt_data):
        if prompt_path is not None and prompt_data is not None:
            prompt = get_prompt(prompt_path, prompt_data)
            logger.info('\n>>>>>>>>>> User Prompt >>>>>>>>>>\n%s\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>', prompt)
            self.messages.append({"role": "user", "content": prompt})

    def get_additional_info_from_user(self, function_calls: FunctionCallSet = None):
        """
        Asks user if he wants to make any changes to last message in conversation.

        Args:
            function_calls: Optional function calls to be included in the message.

        Returns:
            The response from the agent OR None if user didn't ask for change.
        """
        llm_response = None
        while True:
            print(color_yellow(
                "Please check this message and say what needs to be changed. If everything is ok just press ENTER", ))
            changes = ask_user(self.agent.project, self.messages[-1]['content'], require_some_input=False)
            if changes.lower() == '':
                break

            llm_response = self.send_message('utils/update.prompt',
                                             {'changes': changes},
                                             function_calls)

        logger.info('Getting additional info from user done')
        return llm_response



from peewee import TextField
from database.models.components.progress_step import ProgressStep


class ProjectDescription(ProgressStep):
    prompt = TextField()
    summary = TextField()

    class Meta:
        table_name = 'project_description'



from peewee import Model, UUIDField, DateTimeField
from datetime import datetime
from uuid import uuid4

from database.config import DATABASE_TYPE
from database.connection.postgres import get_postgres_database
from database.connection.sqlite import get_sqlite_database


# Establish connection to the database
if DATABASE_TYPE == "postgres":
    database = get_postgres_database()
else:
    database = get_sqlite_database()


class BaseModel(Model):
    id = UUIDField(primary_key=True, default=uuid4)
    created_at = DateTimeField(default=datetime.now)
    updated_at = DateTimeField(default=datetime.now)

    class Meta:
        database = database



import builtins
from json import JSONDecodeError
import os

import pytest
from unittest.mock import call, patch, Mock
from dotenv import load_dotenv
from jsonschema import ValidationError
from const.function_calls import ARCHITECTURE, DEVELOPMENT_PLAN
from helpers.AgentConvo import AgentConvo
from helpers.Project import Project
from helpers.agents.Architect import Architect
from helpers.agents.TechLead import TechLead
from utils.function_calling import parse_agent_response, FunctionType
from test.test_utils import assert_non_empty_string
from test.mock_questionary import MockQuestionary
from utils.llm_connection import create_gpt_chat_completion, stream_gpt_completion, \
    assert_json_response, assert_json_schema, clean_json_response, retry_on_exception
from main import get_custom_print

load_dotenv()
os.environ.pop("AUTOFIX_FILE_PATHS", None)


def test_clean_json_response_True_False():
    # Given a JSON response with Title Case True and False
    response = '''
```json
{
    "steps": [
        {
            "type": "command",
            "command": {
                "command": "git init",
                "daemon": False,
                "timeout": 3000,
                "boolean": False
            },
            "another_True": True,
            "check_if_fixed": True
        }
    ]
}
```
'''

    # When
    response = clean_json_response(response)

    # Then the markdown is removed
    assert response.startswith('{')
    assert response.endswith('}')
    # And the booleans are converted to lowercase
    assert '"daemon":false,' in response
    assert '"boolean":false' in response
    assert '"another_True":true,' in response
    assert '"check_if_fixed":true' in response


def test_clean_json_response_boolean_in_python():
    # Given a JSON response with Python booleans in a content string
    response = '''
{
    "type": "code_change",
    "code_change": {
        "name": "main.py",
        "path": "./main.py",
        "content": "json = {'is_true': True,\\n 'is_false': False}"
    }
}'''

    # When
    response = clean_json_response(response)

    # Then the content string is left untouched
    assert '"content": "json = {\'is_true\': True,\\n \'is_false\': False}"' in response


@patch('utils.llm_connection.styled_text', return_value='')
class TestRetryOnException:
    def setup_method(self):
        self.function: FunctionType = {
            'name': 'test',
            'description': 'test schema',
            'parameters': {
                'type': 'object',
                'properties': {
                    'foo': {'type': 'string'},
                    'boolean': {'type': 'boolean'},
                    'items': {'type': 'array'}
                },
                'required': ['foo']
            }
        }

    def _create_wrapped_function(self, json_responses: list[str]):
        project = Project({'app_id': 'test-app'})
        args = {}, 'test', project

        def retryable_assert_json_schema(data, _req_type, _project):
            json_string = json_responses.pop(0)
            if 'function_buffer' in data:
                json_string = data['function_buffer'] + json_string
            assert_json_schema(json_string, [self.function])
            return json_string

        return retry_on_exception(retryable_assert_json_schema), args

    def test_incomplete_value_string(self, mock_styled_text):
        # Given incomplete JSON
        wrapper, args = self._create_wrapped_function(['{"foo": "bar', '"}'])

        # When
        response = wrapper(*args)

        # Then should tell the LLM the JSON response is incomplete and to continue
        # 'Unterminated string starting at'
        assert response == '{"foo": "bar"}'
        assert 'function_error' not in args[0]
        # And the user should not need to be notified
        assert mock_styled_text.call_count == 0

    def test_incomplete_key(self, mock_styled_text):
        # Given invalid JSON boolean
        wrapper, args = self._create_wrapped_function([
            '{"foo',
            '": "bar"}'
        ])

        # When
        response = wrapper(*args)

        # Then should tell the LLM the JSON response is incomplete and to continue
        # 'Unterminated string starting at: line 1 column 2 (char 1)'
        assert response == '{"foo": "bar"}'
        assert 'function_error' not in args[0]
        # And the user should not need to be notified
        assert mock_styled_text.call_count == 0

    def test_incomplete_value_missing(self, mock_styled_text):
        # Given invalid JSON boolean
        wrapper, args = self._create_wrapped_function([
            '{"foo":',
            ' "bar"}'
        ])

        # When
        response = wrapper(*args)

        # Then should tell the LLM the JSON response is incomplete and to continue
        # 'Expecting value: line 1 column 8 (char 7)'
        assert response == '{"foo": "bar"}'
        assert 'function_error' not in args[0]
        # And the user should not need to be notified
        assert mock_styled_text.call_count == 0

    def test_invalid_boolean(self, mock_styled_text):
        # Given invalid JSON boolean
        wrapper, args = self._create_wrapped_function([
            '{"foo": "bar", "boolean": True}',
            '{"foo": "bar", "boolean": True}',
            '{"foo": "bar", "boolean": True}',
            '{"foo": "bar", "boolean": true}',
        ])

        # When
        response = wrapper(*args)

        # Then should tell the LLM there is an error in the JSON response
        # 'Expecting value: line 1 column 13 (char 12)'
        assert response == '{"foo": "bar", "boolean": true}'
        assert args[0]['function_error'] == 'Invalid value: `True`'
        assert 'function_buffer' not in args[0]
        # And the user should not need to be notified
        assert mock_styled_text.call_count == 1

    def test_invalid_escape(self, mock_styled_text):
        # Given invalid JSON boolean
        wrapper, args = self._create_wrapped_function([
            '{"foo": "\\!"}',
            '{"foo": "\\xBADU"}',
            '{"foo": "\\xd800"}',
            '{"foo": "bar"}',
        ])

        # When
        response = wrapper(*args)

        # Then should tell the LLM there is an error in the JSON response
        # 'Invalid \\escape: line 1 column 10 (char 9)'
        assert response == '{"foo": "bar"}'
        assert len(args[0]['function_error']) > 0
        assert 'function_buffer' not in args[0]
        # And the user should not need to be notified
        assert mock_styled_text.call_count == 1

    def test_incomplete_json_item(self, mock_styled_text):
        # Given incomplete JSON
        wrapper, args = self._create_wrapped_function([
            '{"foo": "bar",',
            ' "boolean"',
            ': true}'])

        # When
        response = wrapper(*args)

        # Then should tell the LLM the JSON response is incomplete and to continue
        # 'Expecting property name enclosed in double quotes: line 1 column 15 (char 14)'
        # "Expecting ':' delimiter: line 1 column 25 (char 24)"
        assert response == '{"foo": "bar", "boolean": true}'
        assert 'function_error' not in args[0]
        # And the user should not need to be notified
        assert mock_styled_text.call_count == 0

    def test_incomplete_json_array(self, mock_styled_text):
        # Given incomplete JSON
        wrapper, args = self._create_wrapped_function([
            '{"foo": "bar", "items": [1, 2, 3, "4"',
            ', 5]}'])

        # When
        response = wrapper(*args)

        # Then should tell the LLM the JSON response is incomplete and to continue
        # "Expecting ',' delimiter: line 1 column 24 (char 23)"
        assert response == '{"foo": "bar", "items": [1, 2, 3, "4", 5]}'
        assert 'function_error' not in args[0]
        # And the user should not need to be notified
        assert mock_styled_text.call_count == 0

    def test_incomplete_then_invalid_by_schema(self, mock_styled_text):
        # Given incomplete JSON
        wrapper, args = self._create_wrapped_function([
            '{"items": [1, 2, 3, "4"',
            ', 5]}',
            # Please try again with a valid JSON object, referring to the previous JSON schema I provided above
            '{"foo": "bar",',
            ' "items": [1, 2, 3, "4"',
            ', 5]}'
        ])

        # When
        response = wrapper(*args)

        # Then should tell the LLM the JSON response is incomplete and to continue
        # "Expecting ',' delimiter: line 1 column 24 (char 23)"
        # "'foo' is a required property"
        assert response == '{"foo": "bar", "items": [1, 2, 3, "4", 5]}'
        assert 'function_error' not in args[0]
        # And the user should not need to be notified
        assert mock_styled_text.call_count == 0

    def test_invalid_boolean_max_retries(self, mock_styled_text):
        # Given invalid JSON boolean
        wrapper, args = self._create_wrapped_function([
            '{"boolean": True, "foo": "bar"}',
            '{"boolean": True,\n "foo": "bar"}',
            '{"boolean": True}',
            '{"boolean": true, "foo": "bar"}',
        ])

        # When
        response = wrapper(*args)

        # Then should tell the LLM there is an error in the JSON response
        assert response == '{"boolean": true, "foo": "bar"}'
        assert args[0]['function_error'] == 'Invalid value: `True`'
        assert mock_styled_text.call_count == 1

    def test_extra_data(self, mock_styled_text):
        # Given invalid JSON boolean
        wrapper, args = self._create_wrapped_function([
            '{"boolean": true, "foo": "bar"}\n I hope that helps',
            '{"boolean": true, "foo": "bar"}\n I hope that helps',
            '{"boolean": true, "foo": "bar"}\n I hope that helps',
            '{"boolean": true, "foo": "bar"}',
        ])

        # When
        response = wrapper(*args)

        # Then should tell the LLM there is an error in the JSON response
        assert response == '{"boolean": true, "foo": "bar"}'
        # assert len(args[0]['function_error']) > 0
        assert args[0]['function_error'] == 'Extra data: line 2 column 2 (char 33)'
        assert mock_styled_text.call_count == 1


class TestSchemaValidation:
    def setup_method(self):
        self.function: FunctionType = {
            'name': 'test',
            'description': 'test schema',
            'parameters': {
                'type': 'object',
                'properties': {'foo': {'type': 'string'}},
                'required': ['foo']
            }
        }

    def test_assert_json_response(self):
        assert assert_json_response('{"foo": "bar"}')
        assert assert_json_response('{\n"foo": "bar"}')
        assert assert_json_response('```\n{"foo": "bar"}')
        assert assert_json_response('```json\n{\n"foo": "bar"}')
        with pytest.raises(ValueError, match='LLM did not respond with JSON'):
            assert assert_json_response('# Foo\n bar')

    def test_assert_json_schema(self):
        # When assert_json_schema is called with valid JSON
        # Then no errors
        assert(assert_json_schema('{"foo": "bar"}', [self.function]))

    def test_assert_json_schema_incomplete(self):
        # When assert_json_schema is called with incomplete JSON
        # Then error is raised
        with pytest.raises(JSONDecodeError):
            assert_json_schema('{"foo": "b', [self.function])

    def test_assert_json_schema_invalid(self):
        # When assert_json_schema is called with invalid JSON
        # Then error is raised
        with pytest.raises(ValidationError, match="1 is not of type 'string'"):
            assert_json_schema('{"foo": 1}', [self.function])

    def test_assert_json_schema_required(self):
        # When assert_json_schema is called with missing required property
        # Then error is raised
        self.function['parameters']['properties']['other'] = {'type': 'string'}
        self.function['parameters']['required'] = ['foo', 'other']

        with pytest.raises(ValidationError, match="'other' is a required property"):
            assert_json_schema('{"foo": "bar"}', [self.function])

    def test_DEVELOPMENT_PLAN(self):
        assert(assert_json_schema('''
{
  "plan": [
    {
      "description": "Set up project structure including creation of necessary directories and files. Initialize Node.js and install necessary libraries such as express and socket.io.",
      "user_review_goal": "Developer should be able to start an empty express server by running `npm start` command without any errors."
    },
    {
      "description": "Create a simple front-end HTML page with CSS and JavaScript that includes input for typing messages and area for displaying messages.",
      "user_review_goal": "Navigating to the root URL (http://localhost:3000) should display the chat front-end with an input box and a message area."
    },
    {
      "description": "Set up socket.io on the back-end to handle websocket connections and broadcasting messages to the clients.",
      "user_review_goal": "By using two different browsers or browser tabs, when one user sends a message from one tab, it should appear in the other user's browser tab in real-time."
    },
    {
      "description": "Integrate front-end with socket.io client to send messages from the input field to the server and display incoming messages in the message area.",
      "user_review_goal": "Typing a message in the chat input and sending it should then display the message in the chat area."
    }
  ]
}
'''.strip(), DEVELOPMENT_PLAN['definitions']))


class TestLlmConnection:
    def setup_method(self):
        builtins.print, ipc_client_instance = get_custom_print({})

    @patch('utils.llm_connection.requests.post')
    @patch('utils.llm_connection.time.sleep')
    def test_rate_limit_error(self, mock_sleep, mock_post, monkeypatch):
        project = Project({'app_id': 'test-app'})

        monkeypatch.setenv('OPENAI_API_KEY', 'secret')

        error_text = '''{
                "error": {
                    "message": "Rate limit reached for 10KTPM-200RPM in organization org-OASFC7k1Ff5IzueeLArhQtnT on tokens per min. Limit: 10000 / min. Please try again in 6ms. Contact us through our help center at help.openai.com if you continue to have issues.",
                    "type": "tokens",
                    "param": null,
                    "code": "rate_limit_exceeded"
                }
            }'''
        content = 'DONE'
        success_text = '{"id": "gen-123", "choices": [{"index": 0, "delta": {"role": "assistant", "content": "' + content + '"}}]}'

        error_response = Mock()
        error_response.status_code = 429
        error_response.text = error_text

        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.iter_lines.return_value = [success_text.encode('utf-8')]

        mock_post.side_effect = [error_response, error_response, error_response, error_response, error_response,
                                 error_response, error_response, error_response, error_response, error_response,
                                 error_response, error_response, mock_response]
        wrapper = retry_on_exception(stream_gpt_completion)
        data = {
            'model': 'gpt-4',
            'messages': [{'role': 'user', 'content': 'testing'}]
        }

        # When
        response = wrapper(data, 'test', project)

        # Then
        assert response == {'text': 'DONE'}
        # assert mock_sleep.call_count == 9
        assert mock_sleep.call_args_list == [call(0.006), call(0.012), call(0.024), call(0.048), call(0.096),
                                             call(0.192), call(0.384), call(0.768), call(1.536), call(3.072),
                                             call(6.144), call(6.144)]
        # mock_sleep.call

    @patch('utils.llm_connection.requests.post')
    def test_stream_gpt_completion(self, mock_post, monkeypatch):
        project = Project({'app_id': 'test-app'})

        # Given streaming JSON response
        monkeypatch.setenv('OPENAI_API_KEY', 'secret')
        deltas = ['{', '\\n',
                  '  \\"foo\\": \\"bar\\",', '\\n',
                  '  \\"prompt\\": \\"Hello\\",', '\\n',
                  '  \\"choices\\": []', '\\n',
                  '}']
        lines_to_yield = [
            ('{"id": "gen-123", "choices": [{"index": 0, "delta": {"role": "assistant", "content": "' + delta + '"}}]}')
            .encode('utf-8')
            for delta in deltas
        ]
        lines_to_yield.insert(1, b': OPENROUTER PROCESSING')  # Simulate OpenRoute keep-alive pings
        mock_response = Mock()
        mock_response.status_code = 200
        mock_response.iter_lines.return_value = lines_to_yield

        mock_post.return_value = mock_response

        with patch('utils.llm_connection.requests.post', return_value=mock_response):
            # When
            response = stream_gpt_completion({
                'model': 'gpt-4',
                'messages': [],
            }, '', project)

            # Then
            assert response == {'text': '{\n  "foo": "bar",\n  "prompt": "Hello",\n  "choices": []\n}'}


    @pytest.mark.uses_tokens
    @pytest.mark.parametrize('endpoint, model', [
        ('OPENAI', 'gpt-4'),                                 # role: system
        ('OPENROUTER', 'openai/gpt-3.5-turbo'),              # role: user
        ('OPENROUTER', 'meta-llama/codellama-34b-instruct'), # rule: user, is_llama
        ('OPENROUTER', 'google/palm-2-chat-bison'),          # role: user/system
        ('OPENROUTER', 'google/palm-2-codechat-bison'),
        ('OPENROUTER', 'anthropic/claude-2'),              # role: user, is_llama
    ])
    def test_chat_completion_Architect(self, endpoint, model, monkeypatch):
        # Given
        monkeypatch.setenv('ENDPOINT', endpoint)
        monkeypatch.setenv('MODEL_NAME', model)
        project = Project({'app_id': 'test-app'})

        agent = Architect(project)
        convo = AgentConvo(agent)
        convo.construct_and_add_message_from_prompt('architecture/technologies.prompt',
                                                        {
                                                            'name': 'Test App',
                                                            'app_summary': '''
The project involves the development of a web-based chat application named "Test_App".
In this application, users can send direct messages to each other.
However, it does not include a group chat functionality.
Multimedia messaging, such as the exchange of images and videos, is not a requirement for this application.
No clear instructions were given for the inclusion of user profile customization features like profile
picture and status updates, as well as a feature for chat history. The project must be developed strictly
as a monolithic application, regardless of any other suggested methods.
The project's specifications are subject to the project manager's discretion, implying a need for
solution-oriented decision-making in areas where precise instructions were not provided.''',
                                                            'app_type': 'web app',
                                                            'user_stories': [
                                                                'User will be able to send direct messages to another user.',
                                                                'User will receive direct messages from other users.',
                                                                'User will view the sent and received messages in a conversation view.',
                                                                'User will select a user to send a direct message.',
                                                                'User will be able to search for users to send direct messages to.',
                                                                'Users can view the online status of other users.',
                                                                'User will be able to log into the application using their credentials.',
                                                                'User will be able to logout from the Test_App.',
                                                                'User will be able to register a new account on Test_App.',
                                                            ]
                                                        })
        function_calls = ARCHITECTURE

        # When
        response = create_gpt_chat_completion(convo.messages, '', project, function_calls=function_calls)

        # Then
        assert convo.messages[0]['content'].startswith('You are an experienced software architect')
        assert convo.messages[1]['content'].startswith('You are working in a software development agency')

        assert response is not None
        response = parse_agent_response(response, function_calls)
        assert 'Node.js' in response['technologies']

    @pytest.mark.uses_tokens
    @pytest.mark.parametrize('endpoint, model', [
        ('OPENAI', 'gpt-4'),
        ('OPENROUTER', 'openai/gpt-3.5-turbo'),
        ('OPENROUTER', 'meta-llama/codellama-34b-instruct'),
        ('OPENROUTER', 'phind/phind-codellama-34b-v2'),
        ('OPENROUTER', 'google/palm-2-chat-bison'),
        ('OPENROUTER', 'google/palm-2-codechat-bison'),
        ('OPENROUTER', 'anthropic/claude-2'),
        ('OPENROUTER', 'mistralai/mistral-7b-instruct')
    ])
    def test_chat_completion_TechLead(self, endpoint, model, monkeypatch):
        # Given
        monkeypatch.setenv('ENDPOINT', endpoint)
        monkeypatch.setenv('MODEL_NAME', model)
        project = Project({'app_id': 'test-app'})

        agent = TechLead(project)
        convo = AgentConvo(agent)
        convo.construct_and_add_message_from_prompt('development/plan.prompt',
                                                    {
                                                        'name': 'Test App',
                                                        'app_summary': '''
    The project entails creating a web-based chat application, tentatively named "chat_app."
This application does not require user authentication or chat history storage.
It solely supports one-on-one messaging, excluding group chats or multimedia sharing like photos, videos, or files.
Additionally, there are no specific requirements for real-time functionality, like live typing indicators or read receipts.
The development of this application will strictly follow a monolithic structure, avoiding the use of microservices, as per the client's demand.
The development process will include the creation of user stories and tasks, based on detailed discussions with the client.''',
                                                        'app_type': 'web app',
                                                        'user_stories': [
            'User Story 1: As a user, I can access the web-based "chat_app" directly without needing to authenticate or log in. Do you want to add anything else? If not, just press ENTER.',
            'User Story 2: As a user, I can start one-on-one conversations with another user on the "chat_app". Do you want to add anything else? If not, just press ENTER.',
            'User Story 3: As a user, I can send and receive messages in real-time within my one-on-one conversation on the "chat_app". Do you want to add anything else? If not, just press ENTER.',
            'User Story 4: As a user, I do not need to worry about deleting or storing my chats because the "chat_app" does not store chat histories. Do you want to add anything else? If not, just press ENTER.',
            'User Story 5: As a user, I will only be able to send text messages, as the "chat_app" does not support any kind of multimedia sharing like photos, videos, or files. Do you want to add anything else? If not, just press ENTER.',
            'User Story 6: As a user, I will not see any live typing indicators or read receipts since the "chat_app" does not provide any additional real-time functionality beyond message exchange. Do you want to add anything else? If not, just press ENTER.',
                                                        ]
                                                    })
        function_calls = DEVELOPMENT_PLAN

        # Retry on bad LLM responses
        # mock_questionary = MockQuestionary(['', '', 'no'])

        # with patch('utils.llm_connection.questionary', mock_questionary):
        # When
        response = create_gpt_chat_completion(convo.messages, '', project, function_calls=function_calls)

        # Then
        assert convo.messages[0]['content'].startswith('You are a tech lead in a software development agency')
        assert convo.messages[1]['content'].startswith('You are working in a software development agency and a project manager and software architect approach you')

        assert response is not None
        response = parse_agent_response(response, function_calls)
        assert_non_empty_string(response['plan'][0]['description'])
        assert_non_empty_string(response['plan'][0]['user_review_goal'])



import ast
import json
from logging import getLogger
from os import getenv, makedirs
from pathlib import Path
import sys
from typing import Any, Optional

from dotenv import load_dotenv

load_dotenv()

log = getLogger(__name__)

AVAILABLE_SETTINGS = [
    "telemetry",
    "openai_api_key",
]


class Settings:
    """
    Application settings

    This object holds all the settings for the application, whether they are
    loaded from the config file, set via environment variables or the command
    line arguments.

    Available settings are listed in the `AVAILABLE_SETTINGS` list.

    This is a singleton object, use it by importing the instance
    directly from the module:

    >>> from utils.settings import settings

    To get a setting:

    >>> settings.openai_api_key

    To get all settings as a dictionary:

    >>> dict(settings)

    To set (update) one setting:

    >>> settings.openai_api_key = "test_key"

    To update multiple settings at once:

    >>> settings.update(openai_api_key="test_key", telemetry=None)

    Note: updating settings will not save them to the config file.
    To do that, use the `loader.save()` method:

    >>> from utils.settings import loader
    >>> loader.save("openai_api_key", "telemetry")

    To see all available settings:

    >>> from utils.settings import AVAILABLE_SETTINGS
    >>> print(AVAILABLE_SETTINGS)
    """

    # Available settings.
    __slots__ = AVAILABLE_SETTINGS

    def __init__(self, **kwargs):
        for key in self.__slots__:
            setattr(self, key, None)

        self.update(**kwargs)

    def __iter__(self):
        for key in self.__slots__:
            yield key, getattr(self, key)

    def update(self, **kwargs):
        """
        Update settings.

        :param kwargs: settings to update (name=value)
        """
        for key, value in kwargs.items():
            try:
                setattr(self, key, value)
            except AttributeError:
                log.warning(f"Ignoring unknown setting: {key}")


class Loader:
    """
    Loader for application settings.

    The app settings are stored in a JSON file in the config directory.

    The config directory is determined by the following rules:
    * If the XDG_CONFIG_HOME environment variable is set (desktop Linux), use that.
    * If the APPDATA environment variable is set (Windows), use that.
    * Otherwise, use the POSIX default ~/.<app-name> (MacOS, server Linux).

    Settings from the config file can be overridden by environment variables
    (all caps) or command line arguments.

    This is a singleton object, use it by importing the instance directly
    from the module:

    >>> from utils.settings import loader

    This will load the settings automatically.

    To get the config file location:

    >>> from utils.settings import config_path
    >>> print(config_path)

    To get the current version of GPT Pilot:

    >>> from utils.settings import version
    >>> print(version)
    """

    APP_NAME = "GPT Pilot"

    def __init__(self, settings: Settings):
        self.config_dir = self.resolve_config_dir()
        self.config_path = self.config_dir / "config.json"
        self.settings = settings

    def load(self):
        """
        Load settings from the config file, environment
        variables and the command-line arguments.
        """
        self.settings.update(**self._load_config())
        self.update_settings_from_env(self.settings)
        self.update_settings_from_args(self.settings)

    @classmethod
    def resolve_config_dir(cls) -> Path:
        """
        Figure out where to store the config file(s).

        :return: path to the desired location config directory

        See the clas docstring for details on how the config directory is
        determined.
        """
        posix_app_name = cls.APP_NAME.replace(" ", "-").lower()

        xdg_config_home = getenv("XDG_CONFIG_HOME")
        if xdg_config_home:
            return Path(xdg_config_home) / Path(posix_app_name)

        if sys.platform == "win32" and getenv("APPDATA"):
            return Path(getenv("APPDATA")) / Path(cls.APP_NAME)

        return Path("~").expanduser() / Path(f".{posix_app_name}")

    def _load_config(self) -> dict[str, Any]:
        """
        Load settings from the config file.

        :returns: dict of settings loaded from the config file

        If the file doesn't exist or there is an error loading the
        config file, an empty settings dict will be returned.

        This is a low-level method used automatically by `Loader.load()`.
        """

        if not self.config_path.exists():
            log.debug(f"Config file not found: {self.config_path}")
            return {}

        log.debug(f"Loading settings from config file: {self.config_path}")
        try:
            with open(self.config_path, "r", encoding="utf-8") as fp:
                return json.load(fp)
        except Exception as e:
            log.error(
                f"Error loading config file {self.config_path}: {e}", exc_info=True
            )
            return {}

    def _save_config(self, config: dict[str, Any]):
        """
        Save provided settings to the config file.

        :param config: dict of settings to save

        This is a low-level method that will overwrite the entire
        config with what's passed in. You should probably use
        `update()` instead.
        """

        if not self.config_dir.exists():
            log.debug(f"Creating config directory: {self.config_dir}")
            makedirs(self.config_dir, exist_ok=True)

        log.debug(f"Saving settings to config file: {self.config_path}")
        with open(self.config_path, "w", encoding="utf-8") as fp:
            json.dump(config, fp, indent=2, sort_keys=True)

    def save(self, *args: list[str]):
        """
        Save one or more settings to the config file, creating it
        if neccessary.

        :param args: list of setting names to set

        This method will update the current config file with
        *ONLY* the settings listed here.

        The reason we don't want to store all settings is that some
        might have been overridden from environment variables or
        command line arguments, that might be temporary.

        We don't want to overwrite the config file with those
        (potentially temporary) values.
        """

        settings_from_config = self._load_config()
        for key in args:
            try:
                value = getattr(self.settings, key)
                settings_from_config[key] = value
            except AttributeError:
                pass

        self._save_config(settings_from_config)

    def update_settings_from_env(self, settings: Settings):
        """
        Update settings from environment variables.

        :param settings: Settings object to update in-place

        Note that environment variable names are hardcoded here,
        because they're not always the same as the setting names.
        """
        # Telemetry (see utils.telemetry)
        telemetry_id = getenv("TELEMETRY_ID")
        telemetry_endpoint = getenv("TELEMETRY_ENDPOINT")

        if settings.telemetry is None and (telemetry_id or telemetry_endpoint):
            settings.telemetry = {}

        if telemetry_id:
            settings.telemetry["id"] = telemetry_id
        if telemetry_endpoint:
            settings.telemetry["endpoint"] = telemetry_endpoint

        # OpenAI API key
        openai_api_key = getenv("OPENAI_API_KEY")
        settings.openai_api_key = openai_api_key

    def update_settings_from_args(self, _settings: Settings):
        """
        Update settings from command line arguments.

        :param settings: Settings object to update in-place

        Note: this is not implemented yet, and currently does nothing.
        """
        # TODO: implement this
        pass


def get_git_commit() -> Optional[str]:
    """
    Return the current git commit (if running from a repo).

    :return: commit hash or None if not running from a git repo
    """
    git_dir = Path(__file__).parent.parent.parent / ".git"
    if not git_dir.is_dir():
        return None

    git_head = git_dir / "HEAD"
    if not git_head.is_file():
        return None

    with open(git_head, "r", encoding="utf-8") as fp:
        ref = fp.read().strip()
        if ref.startswith("ref: "):
            ref = ref[5:]
            with open(git_dir / ref, "r", encoding="utf-8") as fp:
                return fp.read().strip()
        else:
            return ref


def get_package_version() -> str:
    """
    Get package version.

    Note: until we have the packaging set up, this always returns "0.0.0".

    :return: package version as defined in setup.py
    """
    UNKNOWN = "0.0.0"

    setup_file = Path(__file__).parent.parent.parent / "setup.py"
    if not setup_file.is_file():
        return UNKNOWN

    try:
        with open(setup_file, "r", encoding="utf-8") as fp:
            code = ast.parse(fp.read(), filename="setup.py")
            for node in code.body:
                if (
                    isinstance(node, ast.Assign)
                    and len(node.targets) == 1
                    and isinstance(node.targets[0], ast.Name)
                    and node.targets[0].id == "VERSION"
                    and isinstance(node.value, ast.Constant)
                ):
                    return str(node.value.value)
    except:  # noqa
        return UNKNOWN



def get_version() -> str:
    """
    Find and return the current version of GPT Pilot.

    The version string is built from the package version and the current
    git commit hash (if running from a git repo).

    Example: 0.0.0-gitbf01c19

    :return: version string
    """

    version = get_package_version()
    commit = get_git_commit()
    if commit:
        version = version + "-git" + commit[:7]

    return version


version = get_version()
settings = Settings()
loader = Loader(settings)
loader.load()
config_path = loader.config_path


__all__ = ["version", "settings", "loader", "config_path"]



import platform
import uuid
import re

from const.messages import WHEN_USER_DONE
from utils.style import (
    color_green,
    color_green_bold,
    color_red,
    color_red_bold,
    color_yellow_bold,
    color_cyan_bold,
    color_white_bold
)
from helpers.exceptions.TokenLimitError import TokenLimitError
from const.code_execution import MAX_COMMAND_DEBUG_TRIES
from helpers.exceptions.TooDeepRecursionError import TooDeepRecursionError
from helpers.Debugger import Debugger
from utils.questionary import styled_text
from utils.utils import step_already_finished
from helpers.agents.CodeMonkey import CodeMonkey
from logger.logger import logger
from helpers.Agent import Agent
from helpers.AgentConvo import AgentConvo
from utils.utils import should_execute_step, array_of_objects_to_string, generate_app_data
from helpers.cli import run_command_until_success, execute_command_and_check_cli_response, running_processes
from const.function_calls import FILTER_OS_TECHNOLOGIES, EXECUTE_COMMANDS, GET_TEST_TYPE, IMPLEMENT_TASK, COMMAND_TO_RUN
from database.database import save_progress, get_progress_steps, update_app_status
from utils.utils import get_os_info
from utils.telemetry import telemetry

ENVIRONMENT_SETUP_STEP = 'environment_setup'


class Developer(Agent):
    def __init__(self, project):
        super().__init__('full_stack_developer', project)
        self.run_command = None
        self.save_dev_steps = True
        self.debugger = Debugger(self)

    def start_coding(self):
        if not self.project.finished:
            self.project.current_step = 'coding'
            update_app_status(self.project.args['app_id'], self.project.current_step)

            if self.project.skip_steps is None:
                if (not self.project.continuing_project or (
                        'skip_until_dev_step' in self.project.args and self.project.args['skip_until_dev_step'] == '0')):
                    self.project.finish_loading()
                else:
                    self.project.skip_steps = True

        # DEVELOPMENT
        print(color_green_bold("🚀 Now for the actual development...\n"))
        logger.info("Starting to create the actual code...")

        total_tasks = len(self.project.development_plan)
        progress_thresholds = [50]  # Percentages of progress when documentation is created
        documented_thresholds = set()

        for i, dev_task in enumerate(self.project.development_plan):
            current_progress_percent = round((i / total_tasks) * 100, 2)

            for threshold in progress_thresholds:
                if current_progress_percent > threshold and threshold not in documented_thresholds:
                    self.project.technical_writer.document_project(current_progress_percent)
                    documented_thresholds.add(threshold)

            self.project.current_task.start_new_task(dev_task['description'], i + 1)
            self.implement_task(i, dev_task)
            telemetry.inc("num_tasks")

        # DEVELOPMENT END
        self.project.technical_writer.document_project(100)
        self.project.dot_pilot_gpt.chat_log_folder(None)
        if not self.project.finished:
            self.project.finished = True
            update_app_status(self.project.args['app_id'], self.project.current_step)
            message = 'The app is DONE!!! Yay...you can use it now.\n'
            logger.info(message)
            print(color_green_bold(message))
            if not self.project.skip_steps:
                telemetry.set("end_result", "success:initial-project")
                telemetry.send()
        else:
            message = 'Feature complete!\n'
            logger.info(message)
            print(color_green_bold(message))
            if not self.project.skip_steps:
                telemetry.set("end_result", "success:feature")
                telemetry.send()

    def implement_task(self, i, development_task=None):
        print(color_green_bold(f'Implementing task #{i + 1}: ') + color_green(f' {development_task["description"]}\n'))
        self.project.dot_pilot_gpt.chat_log_folder(i + 1)

        convo_dev_task = AgentConvo(self)
        instructions = convo_dev_task.send_message('development/task/breakdown.prompt', {
            "name": self.project.args['name'],
            "app_type": self.project.args['app_type'],
            "app_summary": self.project.project_description,
            "clarifications": self.project.clarifications,
            "user_stories": self.project.user_stories,
            "user_tasks": self.project.user_tasks,
            "array_of_objects_to_string": array_of_objects_to_string,  # TODO check why is this here
            "directory_tree": self.project.get_directory_tree(True),
            "current_task_index": i,
            "development_tasks": self.project.development_plan,
            "files": self.project.get_all_coded_files(),
            "task_type": 'feature' if self.project.finished else 'app'
        })

        instructions_prefix = " ".join(instructions.split()[:5])
        instructions_postfix = " ".join(instructions.split()[-5:])
        response = convo_dev_task.send_message('development/parse_task.prompt', {
            'running_processes': running_processes,
            'os': platform.system(),
            'instructions_prefix': instructions_prefix,
            'instructions_postfix': instructions_postfix,
        }, IMPLEMENT_TASK)
        steps = response['tasks']
        convo_dev_task.remove_last_x_messages(2)

        completed_steps = []

        while True:
            result = self.execute_task(convo_dev_task,
                                       development_task['description'],
                                       steps,
                                       development_task=development_task,
                                       continue_development=True,
                                       is_root_task=True,
                                       continue_from_step=len(completed_steps))

            if result['success']:
                break

            if 'step_index' in result:
                result['os'] = platform.system()
                step_index = result['step_index']
                completed_steps = steps[:step_index + 1]
                result['completed_steps'] = completed_steps
                result['current_step'] = steps[step_index]
                result['next_steps'] = steps[step_index + 1:]
                result['current_step_index'] = step_index

                convo_dev_task.remove_last_x_messages(2)
                # todo before updating task first check if update is needed
                response = convo_dev_task.send_message('development/task/update_task.prompt', result, IMPLEMENT_TASK)
                steps = completed_steps + response['tasks']

            else:
                logger.warning('Testing at end of task failed')
                break

    def step_code_change(self, convo, task_description, step, i, test_after_code_changes):
        if 'code_change_description' in step:
            print(f'Implementing code changes for `{step["code_change_description"]}`')
            code_monkey = CodeMonkey(self.project, self)
            updated_convo = code_monkey.implement_code_changes(convo, step['code_change_description'], step)
            if test_after_code_changes:
                return self.test_code_changes(code_monkey, updated_convo)
            else:
                return {"success": True}

        data = step['save_file']
        self.project.save_file(data)
        return {"success": True}

    def step_modify_file(self, convo, task_description, step, i, test_after_code_changes):
        data = step['modify_file']
        print(f'Updating existing file {data["name"]}: {data["code_change_description"].splitlines()[0]}')
        code_monkey = CodeMonkey(self.project, self)
        code_monkey.implement_code_changes(convo, data['code_change_description'], data)
        return {"success": True}

    def step_command_run(self, convo, step, i, success_with_cli_response=False):
        logger.info('Running command: %s', step['command'])
        data = step['command']
        additional_message = ''  # 'Let\'s start with the step #0:\n' if i == 0 else f'So far, steps { ", ".join(f"#{j}" for j in range(i+1)) } are finished so let\'s do step #{i + 1} now.\n'

        command_id = data['command_id'] if 'command_id' in data else None
        success_message = data['success_message'] if 'success_message' in data else None

        return run_command_until_success(convo, data['command'],
                                         timeout=data['timeout'],
                                         command_id=command_id,
                                         success_message=success_message,
                                         additional_message=additional_message,
                                         success_with_cli_response=success_with_cli_response)

    def step_human_intervention(self, convo, step: dict):
        """
        :param convo:
        :param step: {'human_intervention_description': 'some description'}
        :return: {
          'success': bool
          'user_input': string_from_human
        }
        """
        logger.info('Human intervention needed%s: %s',
                    '' if self.run_command is None else f' for command `{self.run_command}`',
                    step['human_intervention_description'])

        while True:
            human_intervention_description = step['human_intervention_description']

            if not self.run_command:
                self.get_run_command(convo)

            if self.run_command:
                if self.project.check_ipc():
                    print(self.run_command, type='run_command')
                else:
                    human_intervention_description += color_yellow_bold('\n\nIf you want to run the app, just type "r" and press ENTER and that will run `' + self.run_command + '`')

            response = self.project.ask_for_human_intervention('I need human intervention:',
                human_intervention_description,
                cbs={
                    'r': lambda conv: run_command_until_success(conv,
                                                                self.run_command,
                                                                # name the process so the LLM can kill it
                                                                command_id='app',
                                                                # If the app doesn't crash in the first 1st second
                                                                # assume it's good and leave it running.
                                                                # If timeout is None the conversation can't continue
                                                                timeout=None,
                                                                force=True,
                                                                return_cli_response=True)
                },
                convo=convo)

            logger.info('human response: %s', response)
            if 'user_input' not in response:
                continue

            if response['user_input'] == 'continue':
                response['success'] = True
            else:
                response['success'] = self.debugger.debug(convo,
                                                          user_input=response['user_input'],
                                                          issue_description=step['human_intervention_description'])
                # TODO add review

            return response

    def step_test(self, convo, test_command):
        # TODO: don't re-run if it's already running
        should_rerun_command = convo.send_message('dev_ops/should_rerun_command.prompt', test_command)
        if should_rerun_command == 'NO':
            return {'success': True}
        elif should_rerun_command == 'YES':
            logger.info('Re-running test command: %s', test_command)
            cli_response, llm_response = execute_command_and_check_cli_response(convo, test_command)
            logger.info('After running command llm_response: ' + llm_response)
            if llm_response == 'NEEDS_DEBUGGING':
                print(color_red('Got incorrect CLI response:'))
                print(cli_response)
                print(color_red('-------------------'))

            result = {
                'success': llm_response in ["DONE", "SKIP"],
                'cli_response': cli_response
            }
            if cli_response is None:
                result['user_input'] = llm_response
            else:
                result['llm_response'] = llm_response
            return result

    def get_run_command(self, convo):
        llm_response = convo.send_message('development/get_run_command.prompt', {}, COMMAND_TO_RUN)
        self.run_command = llm_response['command']

        # Pattern for triple backtick code block with optional language
        triple_backtick_pattern = r"```(?:\w+\n)?(.*?)```"
        triple_match = re.search(triple_backtick_pattern, self.run_command, re.DOTALL)
        # Pattern for single backtick
        single_backtick_pattern = r"`(.*?)`"
        single_match = re.search(single_backtick_pattern, self.run_command, re.DOTALL)

        if triple_match:
            self.run_command = triple_match.group(1).strip()
        elif single_match:
            self.run_command = single_match.group(1).strip()

    def task_postprocessing(self, convo, development_task, continue_development, task_result, last_branch_name):
        # TODO: why does `run_command` belong to the Developer class, rather than just being passed?
        #       ...It's set by execute_task() -> task_postprocessing(), but that is called by various sources.
        #       What is it at step_human_intervention()?
        self.get_run_command(convo)

        if development_task is not None:
            convo.remove_last_x_messages(2)
            detailed_user_review_goal = convo.send_message('development/define_user_review_goal.prompt', {
                'os': platform.system()
            }, should_log_message=False)
            convo.remove_last_x_messages(2)

        try:
            if continue_development:
                continue_description = detailed_user_review_goal if detailed_user_review_goal is not None else None
                return self.continue_development(convo, last_branch_name, continue_description, development_task)
        except TooDeepRecursionError as e:
            logger.warning('Too deep recursion error. Call dev_help_needed() for human_intervention: %s', e.message)
            return self.dev_help_needed({"type": "human_intervention", "human_intervention_description": e.message})

        return task_result

    def should_retry_step_implementation(self, step, step_implementation_try):
        if step_implementation_try >= MAX_COMMAND_DEBUG_TRIES:
            self.dev_help_needed(step)

        print(color_red_bold('\n--------- LLM Reached Token Limit ----------'))
        print(color_red_bold('Can I retry implementing the entire development step?'))

        answer = ''
        while answer != 'y':
            answer = styled_text(
                self.project,
                'Type y/n'
            )

            logger.info("Retry step implementation? %s", answer)
            if answer == 'n':
                return self.dev_help_needed(step)

        return {"success": False, "retry": True}

    def dev_help_needed(self, step):

        if step['type'] == 'command':
            help_description = (
                        color_red_bold('I tried running the following command but it doesn\'t seem to work:\n\n') +
                        color_white_bold(step['command']['command']) +
                        color_red_bold('\n\nCan you please make it work?'))
        elif step['type'] == 'code_change':
            help_description = step['code_change_description']
        elif step['type'] == 'human_intervention':
            help_description = step['human_intervention_description']

        # TODO remove this
        def extract_substring(s):
            start_idx = s.find('```')
            end_idx = s.find('```', start_idx + 3)

            if start_idx != -1 and end_idx != -1:
                return s[start_idx + 3:end_idx]
            else:
                return s

        # TODO end

        answer = ''
        while answer.lower() != 'continue':
            print(color_red_bold('\n----------------------------- I need your help ------------------------------'))
            print(extract_substring(str(help_description)))
            print(color_red_bold('\n-----------------------------------------------------------------------------'))
            print('continue', type='buttons-only')
            answer = styled_text(
                self.project,
                WHEN_USER_DONE
            )
            logger.info("help needed: %s", answer)

        return {"success": True, "user_input": answer}

    def execute_task(self, convo, task_description, task_steps, test_command=None, reset_convo=True,
                     test_after_code_changes=True, continue_development=False,
                     development_task=None, is_root_task=False, continue_from_step=0):
        function_uuid = str(uuid.uuid4())
        convo.save_branch(function_uuid)

        for (i, step) in enumerate(task_steps):
            # Skip steps before continue_from_step
            if i < continue_from_step:
                continue
            logger.info('---------- execute_task() step #%d: %s', i, step)
            # this if statement is for current way of loading app,
            # once we move to backwards compatibility if statement can be removed
            if not self.project.skip_steps:
                self.project.current_task.inc('steps')

            result = None
            step_implementation_try = 0
            need_to_see_output = 'need_to_see_output' in step and step['need_to_see_output']

            while True:
                try:
                    if reset_convo:
                        convo.load_branch(function_uuid)

                    if step['type'] == 'command':
                        result = self.step_command_run(convo, step, i, success_with_cli_response=need_to_see_output)
                        # if need_to_see_output and 'cli_response' in result:
                        #     result['user_input'] = result['cli_response']

                    elif step['type'] in ['save_file', 'code_change']:
                        result = self.step_code_change(convo, task_description, step, i, test_after_code_changes)

                    elif step['type'] == 'modify_file':
                        result = self.step_modify_file(convo, task_description, step, i, test_after_code_changes)

                    elif step['type'] == 'human_intervention':
                        result = self.step_human_intervention(convo, step)

                    # TODO background_command - if we run commands in background we should have way to kill processes
                    #  and that should be added to function_calls.py DEBUG_STEPS_BREAKDOWN and IMPLEMENT_TASK
                    # elif step['type'] == 'kill_process':
                    #     terminate_named_process(step['kill_process'])
                    #     result = {'success': True}

                    logger.info('  step result: %s', result)

                    if (not result['success']) or (need_to_see_output and result.get("user_input") != "SKIP"):
                        result['step'] = step
                        result['step_index'] = i
                        return result

                    if test_command is not None and ('check_if_fixed' not in step or step['check_if_fixed']):
                        logger.info('check_if_fixed: %s', test_command)
                        result = self.step_test(convo, test_command)
                        logger.info('task result: %s', result)
                        return result

                    break
                except TokenLimitError as e:
                    if is_root_task:
                        response = self.should_retry_step_implementation(step, step_implementation_try)
                        if 'retry' in response:
                            # TODO we can rewind this convo even more
                            convo.load_branch(function_uuid)
                            continue
                        elif 'success' in response:
                            result = response
                            break
                    else:
                        raise e
                except TooDeepRecursionError as e:
                    if is_root_task:
                        result = self.dev_help_needed(step)
                        break
                    else:
                        raise e

        result = {"success": True}  # if all steps are finished, the task has been successfully implemented
        convo.load_branch(function_uuid)
        return self.task_postprocessing(convo, development_task, continue_development, result, function_uuid)

    def continue_development(self, iteration_convo, last_branch_name, continue_description='', development_task=None):
        while True:
            logger.info('Continue development, last_branch_name: %s', last_branch_name)
            if last_branch_name in iteration_convo.branches.keys():  # if user_feedback is not None we create new convo
                iteration_convo.load_branch(last_branch_name)
            user_description = ('Here is a description of what should be working: \n\n' + color_cyan_bold(
                continue_description) + '\n') \
                if continue_description != '' else ''
            user_description = 'Can you check if the app works please? ' + user_description

            if self.run_command:
                if self.project.check_ipc():
                    print(self.run_command, type='run_command')
                else:
                    user_description += color_yellow_bold(
                        '\n\nIf you want to run the app, just type "r" and press ENTER and that will run `' + self.run_command + '`')

            # continue_description = ''
            # TODO: Wait for a specific string in the output or timeout?
            response = self.project.ask_for_human_intervention(
                user_description,
                cbs={'r': lambda convo: run_command_until_success(convo, self.run_command,
                                                                  # name the process so the LLM can kill it
                                                                  command_id='app',
                                                                  # If the app doesn't crash in the first 1st second
                                                                  # assume it's good and leave it running.
                                                                  # If timeout is None the conversation can't continue
                                                                  timeout=None,
                                                                  force=True,
                                                                  return_cli_response=True, is_root_task=True)},
                convo=iteration_convo,
                is_root_task=True)

            logger.info('response: %s', response)
            user_feedback = response['user_input'] if 'user_input' in response else None
            if user_feedback == 'continue':
                return {"success": True, "user_input": user_feedback}

            if user_feedback is not None:
                iteration_convo = AgentConvo(self)
                technologies = (
                    [dep["name"] for dep in self.project.system_dependencies] +
                    [dep["name"] for dep in self.project.package_dependencies]
                )
                iteration_description = iteration_convo.send_message('development/iteration.prompt', {
                    "name": self.project.args['name'],
                    "app_type": self.project.args['app_type'],
                    "app_summary": self.project.project_description,
                    "clarifications": self.project.clarifications,
                    "user_stories": self.project.user_stories,
                    "user_tasks": self.project.user_tasks,
                    "technologies": technologies,
                    "array_of_objects_to_string": array_of_objects_to_string,  # TODO check why is this here
                    "directory_tree": self.project.get_directory_tree(True),
                    "current_task": development_task,
                    "development_tasks": self.project.development_plan,
                    "files": self.project.get_all_coded_files(),
                    "user_input": user_feedback,
                })

                instructions_prefix = " ".join(iteration_description.split()[:5])
                instructions_postfix = " ".join(iteration_description.split()[-5:])

                llm_response = iteration_convo.send_message('development/parse_task.prompt', {
                    'running_processes': running_processes,
                    'os': platform.system(),
                    'instructions_prefix': instructions_prefix,
                    'instructions_postfix': instructions_postfix,
                }, IMPLEMENT_TASK)
                iteration_convo.remove_last_x_messages(2)

                task_steps = llm_response['tasks']
                self.execute_task(iteration_convo, iteration_description, task_steps, is_root_task=True)

    def set_up_environment(self):
        self.project.current_step = ENVIRONMENT_SETUP_STEP
        self.convo_os_specific_tech = AgentConvo(self)

        # If this app_id already did this step, just get all data from DB and don't ask user again
        step = get_progress_steps(self.project.args['app_id'], ENVIRONMENT_SETUP_STEP)
        if step and not should_execute_step(self.project.args['step'], ENVIRONMENT_SETUP_STEP):
            step_already_finished(self.project.args, step)
            return

        print(color_green_bold("Setting up the environment...\n"))
        logger.info("Setting up the environment...")

        for dependency in self.project.system_dependencies:
            if 'description' in dependency:
                dep_text = f"{dependency['name']} ({dependency['description']})"
            else:
                dep_text = dependency['name']

            logger.info('Checking %s', dependency)
            llm_response = self.check_system_dependency(dependency)

            if llm_response == 'DONE':
                print(color_green_bold(f"✅ {dep_text} is available."))
            else:
                if dependency["required_locally"]:
                    remedy_text = "Please install it before proceeding with your app."
                else:
                    remedy_text = "If you want to use it locally, you should install it before proceeding."
                print(color_red_bold(f"❌ {dep_text} is not available. {remedy_text}"))

                print('continue', type='buttons-only')
                styled_text(
                    self.project,
                    "When you're ready to proceed, press ENTER to continue."
                )

        save_progress(self.project.args['app_id'], self.project.current_step, {
            "os_specific_technologies": self.project.system_dependencies,
            "newly_installed_technologies": [],
            "app_data": generate_app_data(self.project.args)
        })

        # ENVIRONMENT SETUP END

    def check_system_dependency(self, dependency):
        convo = AgentConvo(self)
        cli_response, llm_response = execute_command_and_check_cli_response(
            convo,
            {
                "command": dependency["test"],
                "timeout": 10000,
            }
        )
        return llm_response

    def test_code_changes(self, code_monkey, convo):
        return {"success": True}
        logger.info('Testing code changes...')
        llm_response = convo.send_message('development/task/step_check.prompt', {}, GET_TEST_TYPE)
        test_type = llm_response['type']

        if test_type == 'command_test':
            command = llm_response['command']
            return run_command_until_success(convo, command['command'], timeout=command['timeout'])
        elif test_type == 'automated_test':
            # TODO get code monkey to implement the automated test
            pass
        elif test_type == 'manual_test':
            # TODO make the message better
            return_value = {'success': False}
            while not return_value['success']:
                description = llm_response['manual_test_description']
                response = self.project.ask_for_human_intervention(
                    'I need your help. Can you please test if this was successful?',
                    description,
                )

                user_feedback = response['user_input']
                if user_feedback is not None and user_feedback != 'continue':
                    self.debugger.debug(convo, user_input=user_feedback, issue_description=description)
                else:
                    return_value = {'success': True, 'user_input': user_feedback}

            return return_value

    def implement_step(self, convo, step_index, type, description):
        logger.info('Implementing %s step #%d: %s', type, step_index, description)
        # TODO remove hardcoded folder path
        directory_tree = self.project.get_directory_tree(True)
        llm_response = convo.send_message('development/task/next_step.prompt', {
            'finished_steps': [],
            'step_description': description,
            'step_type': type,
            'directory_tree': directory_tree,
            'step_index': step_index
        }, EXECUTE_COMMANDS)

        step_details = llm_response['commands']

        if type == 'COMMAND':
            for cmd in step_details:
                run_command_until_success(convo, cmd['command'], timeout=cmd['timeout'])
        # elif type == 'CODE_CHANGE':
        #     code_changes_details = get_step_code_changes()
        #     # TODO: give to code monkey for implementation
        pass



from utils.utils import step_already_finished
from helpers.Agent import Agent
from utils.style import color_green_bold
from helpers.AgentConvo import AgentConvo

from utils.utils import should_execute_step, generate_app_data
from database.database import save_progress, get_progress_steps, save_feature, get_features_by_app_id
from logger.logger import logger
from const.function_calls import DEVELOPMENT_PLAN

DEVELOPMENT_PLANNING_STEP = 'development_planning'


class TechLead(Agent):
    def __init__(self, project):
        super().__init__('tech_lead', project)
        self.save_dev_steps = False

    def create_development_plan(self):
        self.project.current_step = DEVELOPMENT_PLANNING_STEP
        self.convo_development_plan = AgentConvo(self)

        # If this app_id already did this step, just get all data from DB and don't ask user again
        step = get_progress_steps(self.project.args['app_id'], DEVELOPMENT_PLANNING_STEP)
        if step and not should_execute_step(self.project.args['step'], DEVELOPMENT_PLANNING_STEP):
            step_already_finished(self.project.args, step)
            self.project.development_plan = step['development_plan']
            return

        # DEVELOPMENT PLANNING
        print(color_green_bold("Starting to create the action plan for development...\n"))
        logger.info("Starting to create the action plan for development...")

        # TODO add clarifications
        llm_response = self.convo_development_plan.send_message('development/plan.prompt',
            {
                "name": self.project.args['name'],
                "app_type": self.project.args['app_type'],
                "app_summary": self.project.project_description,
                "clarifications": self.project.clarifications,
                "user_stories": self.project.user_stories,
                "user_tasks": self.project.user_tasks,
                "task_type": 'app',
            }, DEVELOPMENT_PLAN)
        self.project.development_plan = llm_response['plan']

        logger.info('Plan for development is created.')

        save_progress(self.project.args['app_id'], self.project.current_step, {
            "development_plan": self.project.development_plan, "app_data": generate_app_data(self.project.args)
        })

        return

    def create_feature_plan(self, feature_description):
        self.save_dev_steps = True
        self.convo_feature_plan = AgentConvo(self)
        previous_features = get_features_by_app_id(self.project.args['app_id'])

        llm_response = self.convo_feature_plan.send_message('development/feature_plan.prompt',
            {
                "name": self.project.args['name'],
                "app_type": self.project.args['app_type'],
                "app_summary": self.project.project_description,
                "clarifications": self.project.clarifications,
                "user_stories": self.project.user_stories,
                "user_tasks": self.project.user_tasks,
                "technologies": self.project.architecture,
                "directory_tree": self.project.get_directory_tree(True),
                "development_tasks": self.project.development_plan,
                "files": self.project.get_all_coded_files(),
                "previous_features": previous_features,
                "feature_description": feature_description,
                "task_type": 'feature',
            }, DEVELOPMENT_PLAN)

        self.project.development_plan = llm_response['plan']

        logger.info('Plan for feature development is created.')
        return

    def create_feature_summary(self, feature_description):
        self.convo_feature_summary = AgentConvo(self)

        llm_response = self.convo_feature_summary.send_message('development/feature_summary.prompt',
            {
                "name": self.project.args['name'],
                "app_type": self.project.args['app_type'],
                "app_summary": self.project.project_description,
                "feature_description": feature_description,
                "development_tasks": self.project.development_plan,
            })

        self.project.feature_summary = llm_response

        if not self.project.skip_steps:
            save_feature(self.project.args['app_id'],
                         self.project.feature_summary,
                         self.convo_feature_plan.messages,
                         self.project.checkpoints['last_development_step'])

        logger.info('Summary for new feature is created.')
        return



# utils/utils.py

import datetime
import os
import platform
import uuid
import distro
import json
import hashlib
import re
from jinja2 import Environment, FileSystemLoader
from .style import color_green

from const.llm import MAX_QUESTIONS, END_RESPONSE
from const.common import ROLES, STEPS
from logger.logger import logger

prompts_path = os.path.join(os.path.dirname(__file__), '..', 'prompts')
file_loader = FileSystemLoader(prompts_path)
env = Environment(loader=file_loader)


def capitalize_first_word_with_underscores(s):
    # Split the string into words based on underscores.
    words = s.split('_')

    # Capitalize the first word and leave the rest unchanged.
    words[0] = words[0].capitalize()

    # Join the words back into a string with underscores.
    capitalized_string = '_'.join(words)

    return capitalized_string


def get_prompt(prompt_name, data=None):
    if data is None:
        data = {}

    get_prompt_components(data)

    logger.info(f"Getting prompt for {prompt_name}")

    # Load the template
    template = env.get_template(prompt_name)

    # Render the template with the provided data
    output = template.render(data)

    return output


def get_prompt_components(data):
    # This function reads and renders all prompts inside /prompts/components and returns them in dictionary

    # Create an empty dictionary to store the file contents.
    prompts_components = {}
    data.update({
        'MAX_QUESTIONS': MAX_QUESTIONS,
        'END_RESPONSE': END_RESPONSE
    })

    # Create a FileSystemLoader
    prompts_path = os.path.join(os.path.dirname(__file__), '..', 'prompts/components')
    file_loader = FileSystemLoader(prompts_path)

    # Create the Jinja2 environment
    env = Environment(loader=file_loader)

    # Get the list of template names
    template_names = env.list_templates()

    # For each template, load and store its content
    for template_name in template_names:
        # Get the filename without extension as the dictionary key.
        file_key = os.path.splitext(template_name)[0]

        # Load the template and render it with no variables
        file_content = env.get_template(template_name).render(data)

        # Store the file content in the dictionary
        prompts_components[file_key] = file_content

    return data.update(prompts_components)


def get_sys_message(role,args=None):
    """
    :param role: 'product_owner', 'architect', 'dev_ops', 'tech_lead', 'full_stack_developer', 'code_monkey'
    :return: { "role": "system", "content": "You are a {role}... You do..." }
    """
    content = get_prompt(f'system_messages/{role}.prompt',args)

    return {
        "role": "system",
        "content": content
    }


def find_role_from_step(target):
    for role, values in ROLES.items():
        if target in values:
            return role

    return 'product_owner'


def get_os_info():
    os_info = {
        "OS": platform.system(),
        "OS Version": platform.version(),
        "Architecture": platform.architecture()[0],
        "Machine": platform.machine(),
        "Node": platform.node(),
        "Release": platform.release(),
    }

    if os_info["OS"] == "Linux":
        os_info["Distribution"] = distro.name(pretty=True)
    elif os_info["OS"] == "Windows":
        os_info["Win32 Version"] = ' '.join(platform.win32_ver())
    elif os_info["OS"] == "Mac":
        os_info["Mac Version"] = platform.mac_ver()[0]

    # Convert the dictionary to a readable text format
    return array_of_objects_to_string(os_info)


def should_execute_step(arg_step, current_step):
    """
    :param arg_step: `project.args['step']`, may be None
    :param current_step:  The step that would be executed next by the calling method.
    :return: True if `current_step` should be executed.
    """
    arg_step_index = 0 if arg_step is None else STEPS.index(arg_step) if arg_step in STEPS else None
    current_step_index = STEPS.index(current_step) if current_step in STEPS else None

    return arg_step_index is not None and current_step_index is not None and current_step_index >= arg_step_index


def step_already_finished(args, step):
    args.update(step['app_data'])

    message = f"✅  {capitalize_first_word_with_underscores(step['step'])}"
    print(color_green(message))
    logger.info(message)


def generate_app_data(args):
    return {'app_id': args['app_id'], 'app_type': args['app_type']}


def array_of_objects_to_string(array):
    return '\n'.join([f'{key}: {value}' for key, value in array.items()])


def hash_data(data):
    serialized_data = json.dumps(replace_functions(data), sort_keys=True).encode('utf-8')
    return hashlib.sha256(serialized_data).hexdigest()


def replace_functions(obj):
    if isinstance(obj, dict):
        return {k: replace_functions(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [replace_functions(item) for item in obj]
    elif callable(obj):
        return "function"
    else:
        return obj


def fix_json(s):
    s = s.replace('True', 'true')
    s = s.replace('False', 'false')
    # s = s.replace('`', '"')
    return fix_json_newlines(s)


def fix_json_newlines(s):
    pattern = r'("(?:\\\\n|\\.|[^"\\])*")'

    def replace_newlines(match):
        return match.group(1).replace('\n', '\\n')

    return re.sub(pattern, replace_newlines, s)


def clean_filename(filename):
    # Remove invalid characters
    cleaned_filename = re.sub(r'[<>:"/\\|?*]', '', filename)

    # Replace whitespace with underscore
    cleaned_filename = re.sub(r'\s', '_', cleaned_filename)

    return cleaned_filename

def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    elif isinstance(obj, uuid.UUID):
        return str(obj)
    else:
        return str(obj)



from fnmatch import fnmatch
import os.path
from typing import Optional

from const.common import IGNORE_PATHS, IGNORE_SIZE_THRESHOLD


class IgnoreMatcher:
    def __init__(self,
        ignore_paths: Optional[list[str]] = None,
        *,
        root_path: Optional[None] = None,
        ignore_binaries: bool = True,
        ignore_large_files: bool = True,
    ):
        """
        Initialize the IgnoreMatcher object.

        The passed paths (optional) are *added* to the list of
        ignore paths from `const.common.IGNORE_PATHS`.

        :param ignore_paths: List of paths to ignore (optional)
        """
        if ignore_paths is None:
            ignore_paths = []

        self.ignore_paths = ignore_paths + IGNORE_PATHS
        self.ignore_binaries = ignore_binaries
        self.ignore_large_files = ignore_large_files
        self.root_path = root_path

    def ignore(self, path: str) -> bool:
        """
        Check if the given path matches any of the ignore patterns.

        Specified path can be either the full path, or a relative path
        (if root_path was set in the constructor).

        :param path: Path to the file or directory to check
        :return: True if the path matches any of the ignore patterns, False otherwise
        """

        # Turn into absolute (full) path
        if self.root_path and not path.startswith(self.root_path):
            path = os.path.join(self.root_path, path)

        if self.is_in_ignore_list(path):
            return True

        if self.ignore_large_files and self.is_large_file(path):
            return True

        if self.ignore_binaries and self.is_binary(path):
            return True

        return False

    def is_in_ignore_list(self, path: str) -> bool:
        """
        Check if the given path matches any of the ignore patterns.

        :param path: The path to the file or directory to check
        :return: True if the path matches any of the ignore patterns, False otherwise.
        """
        name = os.path.basename(path)
        for pattern in self.ignore_paths:
            if fnmatch(name, pattern):
                return True
        return False

    def is_large_file(self, path: str) -> bool:
        """
        Check if the given file is larger than the threshold.

        :param path: FULL path to the file to check.
        :return: True if the file is larger than the threshold, False otherwise.
        """
        if not os.path.isfile(path):
            return False

        return bool(os.path.getsize(path) > IGNORE_SIZE_THRESHOLD)

    def is_binary(self, path: str) -> bool:
        """
        Check if the given file is binary.

        :param path: FULL path to the file to check.
        :return: True if the file is binary, False otherwise.
        """
        if not os.path.isfile(path):
            return False

        try:
            open(path, "r", encoding="utf-8").read(128*1024)
            return False
        except UnicodeDecodeError:
            return True



from unittest.mock import patch, MagicMock, call
from os.path import normpath, sep
import pytest


from helpers.agents.CodeMonkey import CodeMonkey
from const.function_calls import GET_FILE_TO_MODIFY

@pytest.mark.parametrize(
    ("content", "expected_blocks"),
    [
        ("", []),
        ("no code blocks here", []),
        ("one\n```\ncode block\n```\nwithout CURRENT/NEW tags", []),
        (
            "Change\nCURRENT_CODE:\n```python\nold\n```\nNEW_CODE:\n```\nnew\n```\nEND\n",
            [("old", "new")]
        ),
        (
            "\n".join([
                "Code with markdown blocks in it",
                "CURRENT_CODE:",
                "```markdown",
                "# Title",
                "",
                "```python",
                "print('hello world')",
                "```",
                "Rest of markdown",
                "```",
                "NEW_CODE:",
                "```markdown",
                "# Title",
                "",
                "```python",
                "print('goodbye world')",
                "```",
                "New markdown text here",
                "```",
                "END"
            ]),
            [
                (
                    "# Title\n\n```python\nprint('hello world')\n```\nRest of markdown",
                    "# Title\n\n```python\nprint('goodbye world')\n```\nNew markdown text here",
                )
            ]
        )
    ]
)
def test_get_code_blocks(content, expected_blocks):
    code_monkey = CodeMonkey(None, None)
    assert code_monkey.get_code_blocks(content) == expected_blocks


@pytest.mark.parametrize(
    ("haystack", "needle", "result", "error"),
    [
        ### Oneliner old blocks ###
        # Simple match
        ("first\nsecond\nthird", "second", "first\n@@NEW@@\nthird", None),
        # No match
        ("first\nsecond\nthird", "fourth", None, "not found"),
        # Too many matches on the same indentation level
        ("line\nline", "line", None, "found more than once"),
        # Match, replacement should be indented
        ("first\n    second\nthird", "second", "first\n    @@NEW@@\nthird", None),
        # Too many matches, on different indentation levels
        ("line\n  line", "line", None, "found more than once"),

        ### Multiline old blocks ###
        # Simple match
        ("first\nsecond\nthird", "second\nthird", "first\n@@NEW@@", None),
        # No match
        ("first\nsecond\nthird", "second\n  third", None, "not found"),
        # Too many matches on the same indentation level
        ("a\nb\nc\nd\na\nb", "a\nb", None, "found more than once"),
        # Too many matches on different indentation levels
        ("a\nb\nc\nd\n  a\n  b", "a\nb", None, "found more than once"),
        # Match, replacement should be indented
        ("first\n  second\n  third", "second\nthird", "first\n  @@NEW@@", None),

        ### Multiline with empty lines ###
        # Simple match
        ("first\nsecond\n\nthird", "second\n\nthird", "first\n@@NEW@@", None),
        # Indented match with empty lines also indentend
        ("first\n  second\n  \n  third", "second\n\nthird", "first\n  @@NEW@@", None),
        # Indented match with empty lines not indentend
        ("first\n  second\n\n  third", "second\n\nthird", "first\n  @@NEW@@", None),
    ]
)
def test_replace(haystack, needle, result, error):
    code_monkey = CodeMonkey(None, None)
    if error:
        with pytest.raises(ValueError, match=error):
            code_monkey.replace(haystack, needle, "@@NEW@@")
    else:
        assert code_monkey.replace(haystack, needle, "@@NEW@@") == result


@patch("helpers.agents.CodeMonkey.AgentConvo")
def test_identify_file_to_change(MockAgentConvo):
    mock_convo = MockAgentConvo.return_value
    mock_convo.send_message.return_value = {"file": "file.py"}
    files = CodeMonkey(None, None).identify_file_to_change("some description", [])
    assert files == "file.py"
    mock_convo.send_message.assert_called_once_with(
        "development/identify_files_to_change.prompt",
        {
            "code_changes_description": "some description",
            "files": []
        },
        GET_FILE_TO_MODIFY
    )


def test_dedent():
    old_code = "\n".join([
        "    def foo():",
        "        print('bar')",
    ])
    new_code = "\n".join([
        "  def bar():",
        "      print('foo')",
    ])
    expected_old = "\n".join([
        "  def foo():",
        "      print('bar')",
    ])
    expected_new = "\n".join([
        "def bar():",
        "    print('foo')",
    ])
    result_old, result_new = CodeMonkey.dedent(old_code, new_code)
    assert result_old == expected_old
    assert expected_new == result_new


def test_codemonkey_simple():
    mock_project = MagicMock()
    mock_project.get_all_coded_files.return_value = [
        {
            "path": "",
            "name": "main.py",
            "content": "one to the\nfoo\nto the three to the four"
        },
    ]
    mock_project.get_full_file_path.return_value = ("", normpath("/path/to/main.py"))
    mock_convo = MagicMock()
    mock_convo.send_message.return_value = "## Change\nCURRENT_CODE:\n```\nfoo\n```\nNEW_CODE:\n```\nbar\n```\nEND"

    cm = CodeMonkey(mock_project, None)
    cm.implement_code_changes(
        mock_convo,
        "Modify all references from `foo` to `bar`",
        {
            "path": sep,
            "name": "main.py",
        }
    )

    mock_project.get_all_coded_files.assert_called_once()
    mock_project.get_full_file_path.assert_called_once_with(sep, "main.py")
    mock_convo.send_message.assert_called_once_with(
        "development/implement_changes.prompt", {
        "full_output": False,
        "standalone": False,
        "code_changes_description": "Modify all references from `foo` to `bar`",
        "file_content": "one to the\nfoo\nto the three to the four",
        "file_name": "main.py",
        "files": mock_project.get_all_coded_files.return_value,
    })
    mock_project.save_file.assert_called_once_with({
        "path": sep,
        "name": "main.py",
        "content": "one to the\nbar\nto the three to the four"
    })


@patch("helpers.agents.CodeMonkey.trace_code_event")
def test_codemonkey_retry(trace_code_event):
    file_content = (
        "one to the\nfoo\nto the three to the four\n"
        "the rest of this file is filler so it's big enought not to "
        "trigger the full replace fallback immediately upon the first failure"
    )
    mock_project = MagicMock()
    mock_project.get_all_coded_files.return_value = [
        {
            "path": "",
            "name": "main.py",
            "content": file_content,
        },
    ]
    mock_project.get_full_file_path.return_value = ("", normpath("/path/to/main.py"))
    mock_convo = MagicMock()
    mock_convo.send_message.side_effect = [
        # Incorrect match
        "## Change\nCURRENT_CODE:\n```\ntwo\n```\nNEW_CODE:\n```\nbar\n```\nEND\n",
        # Corrected match on retry
        "Apologies, here is the corrected version. ## Change\nCURRENT_CODE:\n```\n  foo\n```\nNEW_CODE:\n```\n  bar\n```\nEND\n",
    ]

    cm = CodeMonkey(mock_project, None)
    cm.implement_code_changes(
        mock_convo,
        "Modify all references from `foo` to `bar`",
        {
            "path": sep,
            "name": "main.py",
        }
    )

    mock_project.get_all_coded_files.assert_called_once()
    mock_project.get_full_file_path.assert_called_once_with(sep, "main.py")
    mock_convo.send_message.assert_has_calls([
        call(
            "development/implement_changes.prompt", {
                "full_output": False,
                "standalone": False,
                "code_changes_description": "Modify all references from `foo` to `bar`",
                "file_content": file_content,
                "file_name": "main.py",
                "files": mock_project.get_all_coded_files.return_value,
            }
        ),
        call(
            "utils/llm_response_error.prompt", {
                "error": (
                    "Error in change 1:\n"
                    "Old code block not found in the original file:\n```\ntwo\n```\n"
                    "Old block *MUST* contain the exact same text (including indentation, empty lines, etc.) "
                    "as the original file in order to match."
                ),
            }
        )
    ])
    mock_project.save_file.assert_called_once_with({
        "path": sep,
        "name": "main.py",
        "content": file_content.replace("foo", "bar"),
    })
    trace_code_event.assert_called_once_with(
        "codemonkey-file-update-error",
        {
            "error": "replace-errors",
            "llm_response": "## Change\nCURRENT_CODE:\n```\ntwo\n```\nNEW_CODE:\n```\nbar\n```\nEND\n",
            "details": [(1, (
                'Old code block not found in the original file:\n```\ntwo\n```\n'
                'Old block *MUST* contain the exact same text (including indentation, empty lines, etc.) '
                'as the original file in order to match.'
            ))]
        }
    )


@patch("helpers.agents.CodeMonkey.trace_code_event")
def test_codemonkey_partial_retry(trace_code_event):
    file_content = (
        "one to the\nfoo\nto the three to the four\n"
        "the rest of this file is filler so it's big enought not to\n"
        "trigger\nthe full replace fallback immediately upon the first failure"
    )
    mock_project = MagicMock()
    mock_project.get_all_coded_files.return_value = [
        {
            "path": "",
            "name": "main.py",
            "content": file_content,
        },
    ]
    mock_project.get_full_file_path.return_value = ("", normpath("/path/to/main.py"))
    mock_convo = MagicMock()
    mock_convo.send_message.side_effect = [
        # Incorrect match
        (
            "## Change 1\nCURRENT_CODE:\n```\ntwo\n```\nNEW_CODE:\n```\nbar\n```\nEND\n"
            "## Change 2\nCURRENT_CODE:\n```\ntrigger\n```\nNEW_CODE:\n```\ncause\n```\nEND\n"
        ),
        "Apologies, here is the corrected version. ## Change 1\nCURRENT_CODE:\n```\n  foo\n```\nNEW_CODE:\n```\n  bar\n```\nEND\n",
    ]

    cm = CodeMonkey(mock_project, None)
    cm.implement_code_changes(
        mock_convo,
        "Modify all references from `foo` to `bar` and `trigger` to `cause`",
        {
            "path": sep,
            "name": "main.py",
        }
    )

    mock_project.get_all_coded_files.assert_called_once()
    mock_project.get_full_file_path.assert_called_once_with(sep, "main.py")
    mock_convo.send_message.assert_has_calls([
        call(
            "development/implement_changes.prompt", {
                "full_output": False,
                "standalone": False,
                "code_changes_description": "Modify all references from `foo` to `bar` and `trigger` to `cause`",
                "file_content": file_content,
                "file_name": "main.py",
                "files": mock_project.get_all_coded_files.return_value,
            }
        ),
        call(
            "utils/llm_response_error.prompt", {
                "error": (
                    "Some changes were applied, but these failed:\n"
                    "Error in change 1:\n"
                    "Old code block not found in the original file:\n```\ntwo\n```\n"
                    "Old block *MUST* contain the exact same text (including indentation, empty lines, etc.) "
                    "as the original file in order to match.\n"
                    "Please fix the errors and try again (only output the blocks that failed to update, not all of them)."
                ),
            }
        )
    ])
    mock_project.save_file.assert_called_once_with({
        "path": sep,
        "name": "main.py",
        "content": file_content.replace("foo", "bar").replace("trigger", "cause")
    })
    trace_code_event.assert_called_once_with(
        "codemonkey-file-update-error",
        {
            "error": "replace-errors",
            "llm_response": (
                "## Change 1\nCURRENT_CODE:\n```\ntwo\n```\nNEW_CODE:\n```\nbar\n```\nEND\n"
                "## Change 2\nCURRENT_CODE:\n```\ntrigger\n```\nNEW_CODE:\n```\ncause\n```\nEND\n"
            ),
            "details": [(1, (
                'Old code block not found in the original file:\n```\ntwo\n```\n'
                'Old block *MUST* contain the exact same text (including indentation, empty lines, etc.) '
                'as the original file in order to match.'
            ))]
        }
    )


@patch("helpers.agents.CodeMonkey.trace_code_event")
def test_codemonkey_fallback(trace_code_event):
    mock_project = MagicMock()
    mock_project.get_all_coded_files.return_value = [
        {
            "path": "",
            "name": "main.py",
            "content": "one to the\nfoo\nto the three to the four"
        },
    ]
    mock_project.get_full_file_path.return_value = ("", normpath("/path/to/main.py"))
    mock_convo = MagicMock()
    mock_convo.send_message.side_effect = [
        # Incorrect match (END within block), will cause immediate fallback because of short file
        "1 ## Change\nCURRENT_CODE:\n```\nfoo\n```\nNEW_CODE:\n```\nbar\nEND\n```\n",
        # Fallback returns entire new file
        "```\none to the\nbar\nto the three to the four\n```\n",
    ]

    cm = CodeMonkey(mock_project, None)
    cm.implement_code_changes(
        mock_convo,
        "Modify all references from `foo` to `bar`",
        {
            "path": sep,
            "name": "main.py",
        }
    )

    mock_project.get_all_coded_files.assert_called_once()
    mock_project.get_full_file_path.assert_called_once_with(sep, "main.py")
    mock_convo.send_message.assert_has_calls([
        call(
            "development/implement_changes.prompt", {
                "full_output": False,
                "standalone": False,
                "code_changes_description": "Modify all references from `foo` to `bar`",
                "file_content": "one to the\nfoo\nto the three to the four",
                "file_name": "main.py",
                "files": mock_project.get_all_coded_files.return_value,
            }
        ),
        call(
            'development/implement_changes.prompt', {
                "full_output": True,
                "standalone": False,
                "code_changes_description": "Modify all references from `foo` to `bar`",
                "file_content": "one to the\nfoo\nto the three to the four",
                "file_name": "main.py",
                "files": mock_project.get_all_coded_files.return_value,
            }
        )
    ])
    mock_project.save_file.assert_called_once_with({
        "path": sep,
        "name": "main.py",
        "content": "one to the\nbar\nto the three to the four"
    })
    trace_code_event.assert_has_calls([
        call(
            'codemonkey-file-update-error',
            {
                'error': 'error-parsing-blocks',
                'llm_response': '1 ## Change\nCURRENT_CODE:\n```\nfoo\n```\nNEW_CODE:\n```\nbar\nEND\n```\n'
            }
        ),
        call(
            'codemonkey-file-update-error',
            {
                'error': 'fallback-complete-replace',
                'llm_response': '1 ## Change\nCURRENT_CODE:\n```\nfoo\n```\nNEW_CODE:\n```\nbar\nEND\n```\n'
            }
        )
    ])


@patch("helpers.agents.CodeMonkey.trace_code_event")
@patch("helpers.agents.CodeMonkey.get_file_contents")
@patch("helpers.agents.CodeMonkey.AgentConvo")
def test_codemonkey_implement_changes_after_debugging(MockAgentConvo, mock_get_file_contents, trace_code_event):
    """
    Test that the flow to figure out files that need to be changed
    (which happens after debugging where we only have a description of the
    changes needed, not file name).

    Also test standalone conversation (though that's not happening after debugging).
    """
    mock_project = MagicMock()
    mock_project.get_all_coded_files.return_value = []
    mock_project.get_full_file_path.return_value = ("", "/path/to/main.py")
    mock_convo = MockAgentConvo.return_value
    mock_convo.send_message.return_value = "## Change\nCURRENT_CODE:\n```\nfoo\n```\nNEW_CODE:\n```\nbar\n```\nEND"
    mock_get_file_contents.return_value = {
        "name": "main.py",
        "path": "",
        "content": "one to the\nfoo\nto the three to the four",
        "full_path": "/path/to/main.py",
    }

    cm = CodeMonkey(mock_project, None)
    with patch.object(cm, "identify_file_to_change") as mock_identify_file_to_change:
        mock_identify_file_to_change.return_value = "/main.py"
        cm.implement_code_changes(
            None,
            "Modify all references from `foo` to `bar`",
            {},
        )

    MockAgentConvo.assert_called_once_with(cm)
    mock_project.get_all_coded_files.assert_called_once()
    mock_project.get_full_file_path.assert_called_once_with("/", "main.py")
    mock_convo.send_message.assert_called_once_with(
        "development/implement_changes.prompt", {
        "full_output": False,
        "standalone": True,
        "code_changes_description": "Modify all references from `foo` to `bar`",
        "file_content": "one to the\nfoo\nto the three to the four",
        "file_name": "main.py",
        "files": mock_project.get_all_coded_files.return_value,
    })
    mock_project.save_file.assert_called_once_with({
        "path": "/",
        "name": "main.py",
        "content": "one to the\nbar\nto the three to the four"
    })
    trace_code_event.assert_not_called()


@patch("helpers.agents.CodeMonkey.trace_code_event")
@patch("helpers.agents.CodeMonkey.get_file_contents")
def test_codemonkey_original_file_not_found(mock_get_file_contents, _trace_code_event):
    mock_project = MagicMock()
    mock_project.get_all_coded_files.return_value = []
    mock_project.get_full_file_path.return_value = ("", normpath("/path/to/main.py"))
    mock_convo = MagicMock()
    mock_convo.send_message.return_value = "```\none to the\nbar\nto the three to the four\n```\n"
    mock_get_file_contents.side_effect = ValueError("File not found: /path/to/main.py")
    cm = CodeMonkey(mock_project, None)
    cm.implement_code_changes(
        mock_convo,
        "Modify all references from `foo` to `bar`",
        {
            "path": sep,
            "name": "main.py",
        }
    )

    mock_project.get_all_coded_files.assert_called_once()
    mock_project.get_full_file_path.assert_called_once_with(sep, "main.py")
    mock_convo.send_message.assert_called_once_with(
        'development/implement_changes.prompt', {
            "full_output": True,
            "standalone": False,
            "code_changes_description": "Modify all references from `foo` to `bar`",
            "file_content": "",
            "file_name": "main.py",
            "files": mock_project.get_all_coded_files.return_value,
        }
    )
    mock_project.save_file.assert_called_once_with({
        "path": sep,
        "name": "main.py",
        "content": "one to the\nbar\nto the three to the four"
    })



import json
from peewee import TextField


class JSONField(TextField):
    def python_value(self, value):
        if value is not None:
            return json.loads(value)
        return value

    def db_value(self, value):
        if value is not None:
            return json.dumps(value)
        return value



import builtins
import pytest
from unittest.mock import patch
from dotenv import load_dotenv
load_dotenv()

from database.database import create_tables
from helpers.Project import Project
from test.mock_questionary import MockQuestionary
from .main import init, get_custom_print


def test_init():
    # When
    args = init()

    # Then
    for field in ['app_id', 'user_id', 'email']:
        assert args[field] is not None

    for field in ['workspace', 'step']:
        assert args[field] is None


@pytest.mark.slow
@pytest.mark.uses_tokens
@pytest.mark.skip(reason="Uses lots of tokens")
@pytest.mark.parametrize("endpoint, model", [
    ("OPENAI", "gpt-4"),
    ("OPENROUTER", "openai/gpt-3.5-turbo"),
    ("OPENROUTER", "meta-llama/codellama-34b-instruct"),
    ("OPENROUTER", "google/palm-2-chat-bison"),
    ("OPENROUTER", "google/palm-2-codechat-bison"),
    # TODO: See https://github.com/1rgs/jsonformer-claude/blob/main/jsonformer_claude/main.py
    #           https://github.com/guidance-ai/guidance - token healing
    ("OPENROUTER", "anthropic/claude-2"),
])
def test_end_to_end(endpoint, model, monkeypatch):
    # Given
    monkeypatch.setenv('ENDPOINT', endpoint)
    monkeypatch.setenv('MODEL_NAME', model)

    create_tables()
    args = init()
    builtins.print, ipc_client_instance = get_custom_print(args)
    project = Project(args)
    mock_questionary = MockQuestionary([
        'Test App',
        'A web-based chat app',
        # 5 clarifying questions
        'Users can send direct messages to each other but with no group chat functionality',
        'No authentication is required at this stage',
        'Use your best judgement',
        'Use your best judgement',
        'Use your best judgement',
    ])

    # When
    with patch('utils.questionary.questionary', mock_questionary):
        project.start()



from peewee import ForeignKeyField, CharField, BooleanField, DateTimeField
from database.config import DATABASE_TYPE
from database.models.components.base_models import BaseModel
from database.models.app import App
from database.models.components.sqlite_middlewares import JSONField
from playhouse.postgres_ext import BinaryJSONField


class ProgressStep(BaseModel):
    app = ForeignKeyField(App, primary_key=True, on_delete='CASCADE')
    step = CharField()

    if DATABASE_TYPE == 'postgres':
        app_data = BinaryJSONField()
        data = BinaryJSONField(null=True)
        messages = BinaryJSONField(null=True)
    else:
        app_data = JSONField()
        data = JSONField(null=True)
        messages = JSONField(null=True)

    completed = BooleanField(default=False)
    completed_at = DateTimeField(null=True)



# ipc.py
import socket
import json
import time

from utils.utils import json_serial

class IPCClient:
    def __init__(self, port):
        self.ready = False
        client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        print("Connecting to the external process...")
        try:
            client.connect(('localhost', int(port)))
            self.client = client
            print("Connected!")
        except ConnectionRefusedError:
            self.client = None
            print("Connection refused, make sure you started the external process")

    def handle_request(self, message_content):
        print(f"Received request from the external process: {message_content}")
        return message_content  # For demonstration, we're just echoing back the content

    def listen(self):
        if self.client is None:
            print("Not connected to the external process!")
            return

        while True:
            data = self.client.recv(65536)
            message = json.loads(data)

            if message['type'] == 'response':
                # self.client.close()
                return message['content']

    def send(self, data):
        serialized_data = json.dumps(data, default=json_serial)
        data_length = len(serialized_data)
        if self.client is not None:
            self.client.sendall(data_length.to_bytes(4, byteorder='big'))
            self.client.sendall(serialized_data.encode('utf-8'))



from colorama import Fore, Style as ColoramaStyle, init
from enum import Enum
from questionary import Style

# Initialize colorama. Ensures that ANSI codes work on Windows systems.
init(autoreset=True)


class Theme(Enum):
    """
    Enum representing themes, which can be either DARK or LIGHT.
    """
    DARK = 'dark'
    LIGHT = 'light'
    YELLOW = 'yellow'


class ColorName(Enum):
    """
    Enum representing color names and their corresponding ANSI color codes.
    Each color has a normal and a light version, indicated by the two elements in the tuple.
    """
    RED = (Fore.RED, Fore.LIGHTRED_EX)
    GREEN = (Fore.GREEN, Fore.LIGHTGREEN_EX)
    YELLOW = (Fore.YELLOW, Fore.LIGHTYELLOW_EX)
    BLUE = (Fore.BLUE, Fore.LIGHTBLUE_EX)
    CYAN = (Fore.CYAN, Fore.LIGHTCYAN_EX)
    WHITE = (Fore.WHITE, Fore.LIGHTWHITE_EX)


THEME_STYLES = {
    # Style configurations for DARK theme
    Theme.DARK: Style.from_dict({
        'question': '#FFFFFF bold',  # the color and style of the question - White
        'answer': '#FF910A bold',  # the color and style of the answer - Dark Orange / Pumpkin
        'pointer': '#FF4500 bold',  # the color and style of the pointer - Orange Red
        'highlighted': '#63CD91 bold',  # the color and style of the highlighted option - Medium Aquamarine
        'instruction': '#FFFF00 bold'  # the color and style of the instruction - Yellow
    }),
    # Style configurations for LIGHT theme
    Theme.LIGHT: Style.from_dict({
        'question': '#000000 bold',  # the color and style of the question - Black
        'answer': '#FFB74D bold',  # the color and style of the answer - Light Orange
        'pointer': '#FF7043 bold',  # the color and style of the pointer - Light Red
        'highlighted': '#AED581 bold',  # the color and style of the highlighted option - Light Green
        'instruction': '#757575 bold'  # the color and style of the instruction - Grey
    }),
    # Style configurations for LIGHT theme
    Theme.YELLOW: Style.from_dict({
        'question': '#FFFF00 bold',  # the color and style of the question - Black
        'answer': '#FFB74D bold',  # the color and style of the answer - Light Orange
        'pointer': '#FF7043 bold',  # the color and style of the pointer - Light Red
    })
}


class ThemeStyle:
    """
    Class that provides style configurations for DARK and LIGHT themes.
    """

    def __init__(self, theme):
        """
        Initializes a ThemeStyle instance.

        Args:
            theme (Theme): An enum member indicating the theme to use.
        """
        self.theme = theme

    def get_style(self):
        """
        Returns the Style configuration for the current theme.

        Returns:
            questionary.Style: The Style instance for the current theme.
        """
        return THEME_STYLES[self.theme]


class StyleConfig:
    """
    Class to manage the application's style and color configurations.
    """

    def __init__(self, theme: Theme = Theme.DARK):
        """
        Initializes a StyleConfig instance.

        Args:
            theme (Theme, optional): The initial theme to use. Defaults to Theme.DARK.
        """
        self.theme_style = ThemeStyle(theme)
        self.theme = theme

    def get_style(self):
        """
        Retrieves the Style configuration from the theme_style instance.

        Returns:
            questionary.Style: The Style configuration.
        """
        return self.theme_style.get_style()

    def get_color(self, color_name: ColorName):
        """
        Retrieves the ANSI color code for the provided color_name, taking into account the current theme.

        Args:
            color_name (ColorName): Enum member indicating the desired color.

        Returns:
            str: The ANSI color code.
        """
        return color_name.value[self.theme == Theme.LIGHT]

    def set_theme(self, theme: Theme):
        """
        Updates the theme of both the StyleConfig and its theme_style instance.

        Args:
            theme (Theme): Enum member indicating the new theme.
        """
        self.theme = theme
        self.theme_style.theme = theme


def get_color_function(color_name: ColorName, bold: bool = False):
    """
    Returns a function that colorizes text using the provided color_name and optionally makes it bold.

    Args:
        color_name (ColorName): Enum member indicating the color to use.
        bold (bool, optional): If True, the returned function will bold text. Defaults to False.

    Returns:
        Callable[[str], str]: A function that takes a string and returns it colorized.
    """

    def color_func(text: str) -> str:
        """
        Colorizes the input text using the color and boldness provided when `get_color_function` was called.

        Args:
            text (str): The text to colorize.

        Returns:
            str: The colorized text.
        """
        color = style_config.get_color(color_name)
        style = ColoramaStyle.BRIGHT if bold else ""
        reset = ColoramaStyle.RESET_ALL  # Reset code to reset the color
        return f'{color}{style}{text}{reset}'

    return color_func


style_config = StyleConfig()

# Dynamically generate color functions
color_red = get_color_function(ColorName.RED)
color_red_bold = get_color_function(ColorName.RED, True)
color_green = get_color_function(ColorName.GREEN)
color_green_bold = get_color_function(ColorName.GREEN, True)
color_yellow = get_color_function(ColorName.YELLOW)
color_yellow_bold = get_color_function(ColorName.YELLOW, True)
color_blue = get_color_function(ColorName.BLUE)
color_blue_bold = get_color_function(ColorName.BLUE, True)
color_cyan = get_color_function(ColorName.CYAN)
color_cyan_bold = get_color_function(ColorName.CYAN, True)
color_white = get_color_function(ColorName.WHITE)
color_white_bold = get_color_function(ColorName.WHITE, True)



class MockQuestionary:
    def __init__(self, answers=None, initial_state='project_description'):
        if answers is None:
            answers = []
        self.answers = iter(answers)
        self.state = initial_state

    class Style:
        def __init__(self, *args, **kwargs):
            pass

    def text(self, question: str, style=None):
        print('AI: ' + question)
        if question.startswith('User Story'):
            self.state = 'user_stories'
        elif question.endswith('write "DONE"'):
            self.state = 'DONE'
        return self

    def ask(self):
        return self.unsafe_ask()

    def unsafe_ask(self):
        if self.state == 'user_stories':
            answer = ''
        elif self.state == 'DONE':
            answer = 'DONE'
        else:  # if self.state == 'project_description':
            answer = next(self.answers, '')

        print('User:', answer)
        return answer



from .utils import should_execute_step


class TestShouldExecuteStep:
    def test_no_step_arg(self):
        assert should_execute_step(None, 'project_description') is True
        assert should_execute_step(None, 'architecture') is True
        assert should_execute_step(None, 'coding') is True

    def test_skip_step(self):
        assert should_execute_step('architecture', 'project_description') is False
        assert should_execute_step('architecture', 'architecture') is True
        assert should_execute_step('architecture', 'coding') is True

    def test_unknown_step(self):
        assert should_execute_step('architecture', 'unknown') is False
        assert should_execute_step('unknown', 'project_description') is False
        assert should_execute_step('unknown', None) is False
        assert should_execute_step(None, None) is False



import pytest
from unittest.mock import patch

from helpers.AgentConvo import AgentConvo
from helpers.agents import Developer
from .utils import create_project
from helpers.cli import terminate_running_processes
from test.mock_questionary import MockQuestionary


@pytest.mark.ux_test
@patch('utils.questionary.get_saved_user_input')
@patch('helpers.cli.get_saved_command_run')
@patch('helpers.AgentConvo.get_saved_development_step')
@patch('helpers.AgentConvo.save_development_step')
def test_continue_development(mock_4, mock_3, mock_2, mock_1):
    # Given
    project = create_project('continue_development', 'hello_world_server')
    # execute_command(project, 'npm install', 13000)

    developer = Developer(project)
    project.developer = developer
    convo = AgentConvo(developer)
    convo.load_branch = lambda last_branch_name: None
    developer.run_command = 'node server.js'

    # Note: uncomment the following 2 lines and indent the remaining lines when debugging without console input
    mock_questionary = MockQuestionary(['r', 'continue'])
    with patch('utils.questionary.questionary', mock_questionary):

        # When
        # `continue_development` calls `run_command_until_success()` if the user types "r"
        developer.continue_development(convo, 'branch_name', 'The web page should say "Hello, World!"')
        print('end of "continue_development" scenario')

    terminate_running_processes()



import builtins
from helpers.ipc import IPCClient
from const.ipc import MESSAGE_TYPE, LOCAL_IGNORE_MESSAGE_TYPES


def get_custom_print(args):
    built_in_print = builtins.print

    def print_to_external_process(*args, **kwargs):
        # message = " ".join(map(str, args))
        message = args[0]

        if 'type' not in kwargs:
            kwargs['type'] = 'verbose'
        elif kwargs['type'] == MESSAGE_TYPE['local']:
            local_print(*args, **kwargs)
            return

        ipc_client_instance.send({
            'type': MESSAGE_TYPE[kwargs['type']],
            'content': message,
        })
        if kwargs['type'] == MESSAGE_TYPE['user_input_request']:
            return ipc_client_instance.listen()

    def local_print(*args, **kwargs):
        message = " ".join(map(str, args))
        if 'type' in kwargs:
            if kwargs['type'] in LOCAL_IGNORE_MESSAGE_TYPES:
                return
            del kwargs['type']

        built_in_print(message, **kwargs)

    ipc_client_instance = None
    if '--external-log-process-port' in args:
        ipc_client_instance = IPCClient(args['--external-log-process-port'])
        return print_to_external_process, ipc_client_instance
    else:
        return local_print, ipc_client_instance
